# History cache for printer jobs
#
# Copyright (C) 2024 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import time
import logging
from asyncio import Lock
from ..common import (
    JobEvent,
    RequestType,
    HistoryFieldData,
    FieldTracker,
    SqlTableDefinition
)

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    Optional,
    Dict,
    List,
    Tuple
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest, UserInfo
    from .database import MoonrakerDatabase as DBComp
    from .job_state import JobState
    from .file_manager.file_manager import FileManager
    from .database import DBProviderWrapper
    Totals = Dict[str, Union[float, int]]
    AuxTotals = List[Dict[str, Any]]

BASE_TOTALS = {
    "total_jobs": 0,
    "total_time": 0.,
    "total_print_time": 0.,
    "total_filament_used": 0.,
    "longest_job": 0.,
    "longest_print": 0.
}
HIST_TABLE = "job_history"
TOTALS_TABLE = "job_totals"

def _create_totals_list(
    job_totals: Dict[str, Any],
    aux_totals: List[Dict[str, Any]],
    instance: str = "default"
) -> List[Tuple[str, str, Any, Any, str]]:
    """
    Returns a list of Tuples formatted for SQL Database insertion.

    Fields of each tuple are in the following order:
        provider, field, maximum, total, instance_id
    """
    totals_list: List[Tuple[str, str, Any, Any, str]] = []
    for key, value in job_totals.items():
        total = value if key.startswith("total_") else None
        maximum = value if total is None else None
        totals_list.append(("history", key, maximum, total, instance))
    for item in aux_totals:
        if not isinstance(item, dict):
            continue
        totals_list.append(
            (
                item["provider"],
                item["field"],
                item["maximum"],
                item["total"],
                instance
            )
        )
    return totals_list

class TotalsSqlDefinition(SqlTableDefinition):
    name = TOTALS_TABLE
    prototype = (
        f"""
        {TOTALS_TABLE} (
            provider TEXT NOT NULL,
            field TEXT NOT NULL,
            maximum REAL,
            total REAL,
            instance_id TEXT NOT NULL,
            PRIMARY KEY (provider, field, instance_id)
        )
        """
    )
    version = 1

    def migrate(self, last_version: int, db_provider: DBProviderWrapper) -> None:
        if last_version == 0:
            # Migrate from "moonraker" namespace to a table
            logging.info("Migrating history totals from moonraker namespace...")
            hist_ns: Dict[str, Any] = db_provider.get_item("moonraker", "history", {})
            job_totals: Dict[str, Any] = hist_ns.get("job_totals", BASE_TOTALS)
            aux_totals: List[Dict[str, Any]] = hist_ns.get("aux_totals", [])
            if not isinstance(job_totals, dict):
                job_totals = dict(BASE_TOTALS)
            if not isinstance(aux_totals, list):
                aux_totals = []
            totals_list = _create_totals_list(job_totals, aux_totals)
            sql_conn = db_provider.connection
            with sql_conn:
                sql_conn.executemany(
                    f"INSERT OR IGNORE INTO {TOTALS_TABLE} VALUES(?, ?, ?, ?, ?)",
                    totals_list
                )
            try:
                db_provider.delete_item("moonraker", "history")
            except Exception:
                pass

class HistorySqlDefinition(SqlTableDefinition):
    name = HIST_TABLE
    prototype = (
        f"""
        {HIST_TABLE} (
            job_id INTEGER PRIMARY KEY ASC,
            user TEXT NOT NULL,
            filename TEXT,
            status TEXT NOT NULL,
            start_time REAL NOT NULL,
            end_time REAL,
            print_duration REAL NOT NULL,
            total_duration REAL NOT NULL,
            filament_used REAL NOT NULL,
            metadata pyjson,
            auxiliary_data pyjson NOT NULL,
            instance_id TEXT NOT NULL
        )
        """
    )
    version = 1

    def _get_entry_item(
        self, entry: Dict[str, Any], name: str, default: Any = 0.
    ) -> Any:
        val = entry.get(name)
        if val is None:
            return default
        return val

    def migrate(self, last_version: int, db_provider: DBProviderWrapper) -> None:
        if last_version == 0:
            conn = db_provider.connection
            for batch in db_provider.iter_namespace("history", 1000):
                conv_vals: List[Tuple[Any, ...]] = []
                entry: Dict[str, Any]
                for key, entry in batch.items():
                    if not isinstance(entry, dict):
                        logging.info(
                            f"History migration, skipping invalid value: {key} {entry}"
                        )
                        continue
                    try:
                        conv_vals.append(
                            (
                                None,
                                self._get_entry_item(entry, "user", "No User"),
                                self._get_entry_item(entry, "filename", "unknown"),
                                self._get_entry_item(entry, "status", "error"),
                                self._get_entry_item(entry, "start_time"),
                                self._get_entry_item(entry, "end_time"),
                                self._get_entry_item(entry, "print_duration"),
                                self._get_entry_item(entry, "total_duration"),
                                self._get_entry_item(entry, "filament_used"),
                                self._get_entry_item(entry, "metadata", {}),
                                self._get_entry_item(entry, "auxiliary_data", []),
                                "default"
                            )
                        )
                    except KeyError:
                        continue
                if not conv_vals:
                    continue
                placeholders = ",".join("?" * len(conv_vals[0]))
                with conn:
                    conn.executemany(
                        f"INSERT INTO {HIST_TABLE} VALUES({placeholders})",
                        conv_vals
                    )
            db_provider.wipe_local_namespace("history")

class History:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.file_manager: FileManager = self.server.lookup_component('file_manager')
        self.request_lock = Lock()
        FieldTracker.class_init(self)
        self.auxiliary_fields: List[HistoryFieldData] = []
        database: DBComp = self.server.lookup_component("database")
        self.history_table = database.register_table(HistorySqlDefinition())
        self.totals_table = database.register_table(TotalsSqlDefinition())
        self.job_totals: Totals = dict(BASE_TOTALS)
        self.aux_totals: AuxTotals = []

        self.server.register_event_handler(
            "server:klippy_disconnect", self._handle_disconnect)
        self.server.register_event_handler(
            "server:klippy_shutdown", self._handle_shutdown)
        self.server.register_event_handler(
            "job_state:state_changed", self._on_job_state_changed)
        self.server.register_event_handler(
            "klippy_apis:job_start_complete", self._on_job_requested)
        self.server.register_notification("history:history_changed")

        self.server.register_endpoint(
            "/server/history/job", RequestType.GET | RequestType.DELETE,
            self._handle_job_request
        )
        self.server.register_endpoint(
            "/server/history/list", RequestType.GET, self._handle_jobs_list
        )
        self.server.register_endpoint(
            "/server/history/totals", RequestType.GET, self._handle_job_totals
        )
        self.server.register_endpoint(
            "/server/history/reset_totals", RequestType.POST,
            self._handle_job_total_reset
        )

        self.current_job: Optional[PrinterJob] = None
        self.current_job_id: Optional[int] = None
        self.job_user: str = "No User"
        self.job_paused: bool = False

    async def component_init(self) -> None:
        # Popluate totals
        valid_aux_totals = [
            (item.provider, item.name) for item in self.auxiliary_fields
            if item.has_totals()
        ]
        cursor = await self.totals_table.execute(f"SELECT * from {TOTALS_TABLE}")
        await cursor.set_arraysize(200)
        for row in await cursor.fetchall():
            provider, field, maximum, total, _ = tuple(row)
            if provider == "history":
                self.job_totals[field] = total if maximum is None else maximum
            elif (provider, field) in valid_aux_totals:
                item = dict(row)
                item.pop("instance_id", None)
                self.aux_totals.append(item)
        # Check for interupted jobs
        cursor = await self.history_table.execute(
            f"SELECT job_id FROM {HIST_TABLE} WHERE status = 'in_progress'"
        )
        interrupted_jobs: List[int] = [row[0] for row in await cursor.fetchall()]
        if interrupted_jobs:
            async with self.history_table as tx:
                await tx.execute(
                    f"UPDATE {HIST_TABLE} SET status = 'interrupted' "
                    "WHERE status = 'in_progress'"
                )
            self.server.add_log_rollover_item(
                "interrupted_history",
                "The following jobs were detected as interrupted: "
                f"{interrupted_jobs}"
            )

    async def _handle_job_request(self,
                                  web_request: WebRequest
                                  ) -> Dict[str, Any]:
        async with self.request_lock:
            req_type = web_request.get_request_type()
            if req_type == RequestType.GET:
                job_id = web_request.get_str("uid")
                cursor = await self.history_table.execute(
                    f"SELECT * FROM {HIST_TABLE} WHERE job_id = ?", (int(job_id, 16),)
                )
                result = await cursor.fetchone()
                if result is None:
                    raise self.server.error(f"Invalid job uid: {job_id}", 404)
                job = dict(result)
                return {"job": self._prep_requested_job(job, job_id)}
            if req_type == RequestType.DELETE:
                all = web_request.get_boolean("all", False)
                if all:
                    cursor = await self.history_table.execute(
                        f"SELECT job_id FROM {HIST_TABLE} WHERE instance_id = ?",
                        ("default",)
                    )
                    await cursor.set_arraysize(1000)
                    deljobs = [f"{row[0]:06X}" for row in await cursor.fetchall()]
                    async with self.history_table as tx:
                        await tx.execute(
                            f"DELETE FROM {HIST_TABLE} WHERE instance_id = ?",
                            ("default",)
                        )
                    return {'deleted_jobs': deljobs}

                job_id = web_request.get_str("uid")
                async with self.history_table as tx:
                    cursor = await tx.execute(
                        f"DELETE FROM {HIST_TABLE} WHERE job_id = ?", (int(job_id, 16),)
                    )
                if cursor.rowcount < 1:
                    raise self.server.error(f"Invalid job uid: {job_id}", 404)
                return {'deleted_jobs': [job_id]}
            raise self.server.error("Invalid Request Method")

    async def _handle_jobs_list(self,
                                web_request: WebRequest
                                ) -> Dict[str, Any]:
        async with self.request_lock:
            before = web_request.get_float("before", -1)
            since = web_request.get_float("since", -1)
            limit = web_request.get_int("limit", 50)
            start = web_request.get_int("start", 0)
            order = web_request.get_str("order", "desc").upper()

            if order not in ["ASC", "DESC"]:
                raise self.server.error(f"Invalid `order` value: {order}", 400)
            # Build SQL Select Statement
            values: List[Any] = ["default"]
            sql_statement = f"SELECT * FROM {HIST_TABLE} WHERE instance_id = ?"
            if before != -1:
                sql_statement += " and end_time < ?"
                values.append(before)
            if since != -1:
                sql_statement += " and start_time > ?"
                values.append(since)
            sql_statement += f" ORDER BY job_id {order}"
            if limit > 0:
                sql_statement += " LIMIT ? OFFSET ?"
                values.append(limit)
                values.append(start)
            cursor = await self.history_table.execute(sql_statement, values)
            await cursor.set_arraysize(1000)
            jobs: List[Dict[str, Any]] = []
            for row in await cursor.fetchall():
                job = dict(row)
                job_id = f"{row['job_id']:06X}"
                jobs.append(self._prep_requested_job(job, job_id))
            return {"count": len(jobs), "jobs": jobs}

    async def _handle_job_totals(
        self, web_request: WebRequest
    ) -> Dict[str, Union[Totals, AuxTotals]]:
        return {
            "job_totals": self.job_totals,
            "auxiliary_totals": self.aux_totals
        }

    async def _handle_job_total_reset(
        self, web_request: WebRequest
    ) -> Dict[str, Union[Totals, AuxTotals]]:
        if self.current_job is not None:
            raise self.server.error("Job in progress, cannot reset totals")
        last_totals = self.job_totals
        self.job_totals = dict(BASE_TOTALS)
        last_aux_totals = self.aux_totals
        self._update_aux_totals(reset=True)
        totals_list = _create_totals_list(self.job_totals, self.aux_totals)
        async with self.totals_table as tx:
            await tx.execute(
                f"DELETE FROM {TOTALS_TABLE} WHERE instance_id = ?", ("default",)
            )
            await tx.executemany(
                f"INSERT INTO {TOTALS_TABLE} VALUES(?, ?, ?, ?, ?)", totals_list
            )
        return {
            "last_totals": last_totals,
            "last_auxiliary_totals": last_aux_totals
        }

    async def _on_job_state_changed(
        self,
        job_event: JobEvent,
        prev_stats: Dict[str, Any],
        new_stats: Dict[str, Any]
    ) -> None:
        self.job_paused = job_event == JobEvent.PAUSED
        if job_event == JobEvent.STARTED:
            if self.current_job is not None:
                # Finish with the previous state
                await self.finish_job("cancelled", prev_stats)
            await self.add_job(PrinterJob(new_stats))
        elif job_event == JobEvent.COMPLETE:
            await self.finish_job("completed", new_stats)
        elif job_event == JobEvent.ERROR:
            await self.finish_job("error", new_stats)
        elif job_event in (JobEvent.CANCELLED, JobEvent.STANDBY):
            # Cancel on "standby" for backward compatibility with
            # `CLEAR_PAUSE/SDCARD_RESET_FILE` workflow
            await self.finish_job("cancelled", prev_stats)

    def _on_job_requested(self, user: Optional[UserInfo]) -> None:
        username = user.username if user is not None else "No User"
        self.job_user = username
        if self.current_job is not None:
            self.current_job.user = username

    async def _handle_shutdown(self) -> None:
        jstate: JobState = self.server.lookup_component("job_state")
        last_ps = jstate.get_last_stats()
        await self.finish_job("klippy_shutdown", last_ps)

    async def _handle_disconnect(self) -> None:
        jstate: JobState = self.server.lookup_component("job_state")
        last_ps = jstate.get_last_stats()
        await self.finish_job("klippy_disconnect", last_ps)

    async def add_job(self, job: PrinterJob) -> None:
        async with self.request_lock:
            self.current_job = job
            self.current_job_id = None
            self.current_job.user = self.job_user
            self.grab_job_metadata()
            for field in self.auxiliary_fields:
                field.tracker.reset()
            self.current_job.set_aux_data(self.auxiliary_fields)
            new_id = await self.save_job(job, None)
            if new_id is None:
                logging.info(f"Error saving job, filename '{job.filename}'")
                return
            self.current_job_id = new_id
            job_id = f"{new_id:06X}"
            self.update_metadata(job_id)
            logging.debug(
                f"History Job Added - Id: {job_id}, File: {job.filename}"
            )
            self.send_history_event("added")

    async def save_job(self, job: PrinterJob, job_id: Optional[int]) -> Optional[int]:
        values: List[Any] = [
            job_id,
            job.user,
            job.filename,
            job.status,
            job.start_time,
            job.end_time,
            job.print_duration,
            job.total_duration,
            job.filament_used,
            job.metadata,
            job.auxiliary_data,
            "default"
        ]
        placeholders = ",".join("?" * len(values))
        async with self.history_table as tx:
            cursor = await tx.execute(
                f"REPLACE INTO {HIST_TABLE} VALUES({placeholders})", values
            )
        return cursor.lastrowid

    async def delete_job(self, job_id: Union[int, str]) -> None:
        if isinstance(job_id, str):
            job_id = int(job_id, 16)
        async with self.history_table as tx:
            tx.execute(
                f"DELETE FROM {HIST_TABLE} WHERE job_id = ?", (job_id,)
            )

    async def finish_job(self, status: str, pstats: Dict[str, Any]) -> None:
        async with self.request_lock:
            if self.current_job is None or self.current_job_id is None:
                self._reset_current_job()
                return
            if (
                pstats.get('filename') != self.current_job.filename or
                pstats.get('total_duration', 0.) < self.current_job.total_duration
            ):
                # Print stats have been reset, do not update this job with them
                pstats = {}
            self.current_job.user = self.job_user
            self.current_job.finish(status, pstats)
            # Regrab metadata incase metadata wasn't parsed yet due to file upload
            self.grab_job_metadata()
            self.current_job.set_aux_data(self.auxiliary_fields)
            job_id = f"{self.current_job_id:06X}"
            await self.save_job(self.current_job, self.current_job_id)
            self.update_metadata(job_id)
            await self._update_job_totals()
            logging.debug(
                f"History Job Finished - Id: {job_id}, "
                f"File: {self.current_job.filename}, "
                f"Status: {status}"
            )
            self.send_history_event("finished")
            self._reset_current_job()

    def _reset_current_job(self) -> None:
        self.current_job = None
        self.current_job_id = None
        self.job_user = "No User"

    async def get_job(
        self, job_id: Union[int, str]
    ) -> Optional[Dict[str, Any]]:
        if isinstance(job_id, str):
            job_id = int(job_id, 16)
        cursor = await self.history_table.execute(
            f"SELECT * FROM {HIST_TABLE} WHERE job_id = ?", (job_id,)
        )
        result = await cursor.fetchone()
        return dict(result) if result is not None else result

    def grab_job_metadata(self) -> None:
        if self.current_job is None:
            return
        filename: str = self.current_job.filename
        mdst = self.file_manager.get_metadata_storage()
        metadata: Dict[str, Any] = mdst.get(filename, {})
        # We don't need to store these fields in the
        # job metadata, as they are redundant
        metadata.pop('print_start_time', None)
        metadata.pop('job_id', None)
        if "thumbnails" in metadata:
            thumb: Dict[str, Any]
            for thumb in metadata['thumbnails']:
                thumb.pop('data', None)
        self.current_job.metadata = metadata

    def update_metadata(self, job_id: str) -> None:
        if self.current_job is None:
            return
        mdst = self.file_manager.get_metadata_storage()
        filename: str = self.current_job.filename
        metadata: Dict[str, Any] = mdst.get(filename, {})
        if metadata:
            # Add the start time and job id to the
            # persistent metadata storage
            metadata.update({
                'print_start_time': self.current_job.get('start_time'),
                'job_id': job_id
            })
            mdst.insert(filename, metadata)

    async def _update_job_totals(self) -> None:
        if self.current_job is None:
            return
        job = self.current_job
        self._accumulate_total("total_jobs", 1)
        self._accumulate_total("total_time", job.total_duration)
        self._accumulate_total("total_print_time", job.print_duration)
        self._accumulate_total("total_filament_used", job.filament_used)
        self._maximize_total("longest_job", job.total_duration)
        self._maximize_total("longest_print", job.print_duration)
        self._update_aux_totals()
        totals_list = _create_totals_list(self.job_totals, self.aux_totals)
        async with self.totals_table as tx:
            await tx.executemany(
                f"REPLACE INTO {TOTALS_TABLE} VALUES(?, ?, ?, ?, ?)", totals_list
            )

    def _accumulate_total(self, field: str, val: Union[int, float]) -> None:
        self.job_totals[field] += val

    def _maximize_total(self, field: str, val: Union[int, float]) -> None:
        self.job_totals[field] = max(self.job_totals[field], val)

    def _update_aux_totals(self, reset: bool = False) -> None:
        last_totals = self.aux_totals
        self.aux_totals = [
            field.get_totals(last_totals, reset)
            for field in self.auxiliary_fields
            if field.has_totals()
        ]

    def send_history_event(self, evt_action: str) -> None:
        if self.current_job is None or self.current_job_id is None:
            return
        job_id = f"{self.current_job_id:06X}"
        job = self._prep_requested_job(self.current_job.get_stats(), job_id)
        self.server.send_event(
            "history:history_changed", {'action': evt_action, 'job': job}
        )

    def _prep_requested_job(
        self, job: Dict[str, Any], job_id: str
    ) -> Dict[str, Any]:
        fm = self.file_manager
        mtime = job.get("metadata", {}).get("modified", None)
        job["exists"] = fm.check_file_exists("gcodes", job['filename'], mtime)
        job["job_id"] = job_id
        job.pop("instance_id", None)
        return job

    def register_auxiliary_field(self, new_field: HistoryFieldData) -> None:
        if new_field.provider == "history":
            raise self.server.error("Provider name 'history' is reserved")
        for field in self.auxiliary_fields:
            if field == new_field:
                raise self.server.error(
                    f"Field {field.name} already registered by "
                    f"provider {field.provider}."
                )
        self.auxiliary_fields.append(new_field)

    def tracking_enabled(self, check_paused: bool) -> bool:
        if self.current_job is None:
            return False
        return not self.job_paused if check_paused else True

    async def on_exit(self) -> None:
        if self.current_job is None:
            return
        jstate: JobState = self.server.lookup_component("job_state")
        last_ps = jstate.get_last_stats()
        await self.finish_job("server_exit", last_ps)

class PrinterJob:
    def __init__(self, data: Dict[str, Any] = {}) -> None:
        self.end_time: Optional[float] = None
        self.filament_used: float = 0
        self.filename: str = ""
        self.metadata: Dict[str, Any] = {}
        self.print_duration: float = 0.
        self.status: str = "in_progress"
        self.start_time = time.time()
        self.total_duration: float = 0.
        self.auxiliary_data: List[Dict[str, Any]] = []
        self.user: str = "No User"
        self.update_from_ps(data)

    def finish(self,
               status: str,
               print_stats: Dict[str, Any] = {}
               ) -> None:
        self.end_time = time.time()
        self.status = status if status is not None else "error"
        self.update_from_ps(print_stats)

    def get(self, name: str) -> Any:
        if not hasattr(self, name):
            return None
        return getattr(self, name)

    def get_stats(self) -> Dict[str, Any]:
        return self.__dict__.copy()

    def set(self, name: str, val: Any) -> None:
        if not hasattr(self, name):
            return
        setattr(self, name, val)

    def set_aux_data(self, fields: List[HistoryFieldData]) -> None:
        self.auxiliary_data = [field.as_dict() for field in fields]

    def update_from_ps(self, data: Dict[str, Any]) -> None:
        for i in data:
            if hasattr(self, i) and data[i] is not None:
                setattr(self, i, data[i])


def load_component(config: ConfigHelper) -> History:
    return History(config)



# SimplyPrint Connection Support
#
# Copyright (C) 2022  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import os
import asyncio
import logging
import time
import pathlib
import base64
import tornado.websocket
from tornado.escape import url_escape
import logging.handlers
import tempfile
from queue import SimpleQueue
from ..loghelper import LocalQueueHandler
from ..common import APITransport, JobEvent, KlippyState, UserInfo
from ..utils import json_wrapper as jsonw

from typing import (
    TYPE_CHECKING,
    Awaitable,
    Optional,
    Dict,
    List,
    Union,
    Any,
    Callable,
)
if TYPE_CHECKING:
    from .application import InternalTransport
    from ..confighelper import ConfigHelper
    from .websockets import WebsocketManager
    from ..common import BaseRemoteConnection
    from tornado.websocket import WebSocketClientConnection
    from .database import MoonrakerDatabase
    from .klippy_apis import KlippyAPI
    from .job_state import JobState
    from .machine import Machine
    from .file_manager.file_manager import FileManager
    from .http_client import HttpClient
    from .power import PrinterPower
    from .announcements import Announcements
    from .webcam import WebcamManager, WebCam
    from .klippy_connection import KlippyConnection

COMPONENT_VERSION = "0.0.1"
SP_VERSION = "0.1"
TEST_ENDPOINT = f"wss://testws.simplyprint.io/{SP_VERSION}/p"
PROD_ENDPOINT = f"wss://ws.simplyprint.io/{SP_VERSION}/p"
# TODO: Increase this time to something greater, perhaps 30 minutes
CONNECTION_ERROR_LOG_TIME = 60.
PRE_SETUP_EVENTS = [
    "connection", "state_change", "shutdown", "machine_data", "firmware",
    "ping"
]

class SimplyPrint(APITransport):
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self._logger = ProtoLogger(config)
        self.eventloop = self.server.get_event_loop()
        self.job_state: JobState
        self.job_state = self.server.lookup_component("job_state")
        self.klippy_apis: KlippyAPI
        self.klippy_apis = self.server.lookup_component("klippy_apis")
        database: MoonrakerDatabase = self.server.lookup_component("database")
        database.register_local_namespace("simplyprint", forbidden=True)
        self.spdb = database.wrap_namespace("simplyprint")
        self.sp_info = self.spdb.as_dict()
        self.is_closing = False
        self.ws: Optional[WebSocketClientConnection] = None
        self.cache = ReportCache()
        ambient = self.sp_info.get("ambient_temp", INITIAL_AMBIENT)
        self.amb_detect = AmbientDetect(config, self, ambient)
        self.layer_detect = LayerDetect()
        self.webcam_stream = WebcamStream(config, self)
        self.print_handler = PrintHandler(self)
        self.last_received_temps: Dict[str, float] = {}
        self.last_err_log_time: float = 0.
        self.last_cpu_update_time: float = 0.
        self.intervals: Dict[str, float] = {
            "job": 1.,
            "temps": 1.,
            "temps_target": .25,
            "cpu": 10.,
            "ai": 0.,
            "ping": 20.,
        }
        self.printer_status: Dict[str, Dict[str, Any]] = {}
        self.heaters: Dict[str, str] = {}
        self.missed_job_events: List[Dict[str, Any]] = []
        self.announce_mutex = asyncio.Lock()
        self.connection_task: Optional[asyncio.Task] = None
        self.reconnect_delay: float = 1.
        self.reconnect_token: Optional[str] = None
        self._last_sp_ping: float = 0.
        self.ping_sp_timer = self.eventloop.register_timer(self._handle_sp_ping)
        self.printer_info_timer = self.eventloop.register_timer(
            self._handle_printer_info_update)
        self._print_request_event: asyncio.Event = asyncio.Event()
        self.next_temp_update_time: float = 0.
        self._last_ping_received: float = 0.
        self.gcode_terminal_enabled: bool = False
        self.connected = False
        self.is_set_up = False
        self.test = config.get("use_test_endpoint", False)
        connect_url = config.get("url", None)
        if connect_url is not None:
            self.connect_url = connect_url
            self.is_set_up = True
        else:
            self._set_ws_url()

        self.power_id: str = ""
        power_id: Optional[str] = config.get("power_device", None)
        if power_id is not None:
            self.power_id = power_id
            if self.power_id.startswith("power "):
                self.power_id = self.power_id[6:]
            if not config.has_section(f"power {self.power_id}"):
                self.power_id = ""
                self.server.add_warning(
                    "Section [simplyprint], option 'power_device': Unable "
                    f"to locate configuration for power device {power_id}"
                )
        else:
            power_pfx = config.get_prefix_sections("power ")
            if len(power_pfx) == 1:
                name = power_pfx[0][6:]
                if "printer" in name.lower():
                    self.power_id = name
        self.filament_sensor: str = ""
        fsensor = config.get("filament_sensor", None)
        fs_prefixes = ["filament_switch_sensor ", "filament_motion_sensor "]
        if fsensor is not None:
            for prefix in fs_prefixes:
                if fsensor.startswith(prefix):
                    self.filament_sensor = fsensor
                    break
            else:
                self.server.add_warning(
                    "Section [simplyprint], option 'filament_sensor': Invalid "
                    f"sensor '{fsensor}', must start with one the following "
                    f"prefixes: {fs_prefixes}"
                )

        # Register State Events
        self.server.register_event_handler(
            "server:klippy_started", self._on_klippy_startup)
        self.server.register_event_handler(
            "server:klippy_ready", self._on_klippy_ready)
        self.server.register_event_handler(
            "server:klippy_shutdown", self._on_klippy_shutdown)
        self.server.register_event_handler(
            "server:klippy_disconnect", self._on_klippy_disconnected)
        self.server.register_event_handler(
            "job_state:state_changed", self._on_job_state_changed)
        self.server.register_event_handler(
            "klippy_apis:pause_requested", self._on_pause_requested)
        self.server.register_event_handler(
            "klippy_apis:resume_requested", self._on_resume_requested)
        self.server.register_event_handler(
            "klippy_apis:cancel_requested", self._on_cancel_requested)
        self.server.register_event_handler(
            "proc_stats:proc_stat_update", self._on_proc_update)
        self.server.register_event_handler(
            "proc_stats:cpu_throttled", self._on_cpu_throttled
        )
        self.server.register_event_handler(
            "websockets:client_identified", self._on_websocket_identified)
        self.server.register_event_handler(
            "websockets:client_removed", self._on_websocket_removed)
        self.server.register_event_handler(
            "server:gcode_response", self._on_gcode_response)
        self.server.register_event_handler(
            "klippy_connection:gcode_received", self._on_gcode_received
        )
        self.server.register_event_handler(
            "power:power_changed", self._on_power_changed
        )

    async def component_init(self) -> None:
        await self.webcam_stream.intialize_url()
        await self.webcam_stream.test_connection()
        self.connection_task = self.eventloop.create_task(self._connect())

    async def _connect(self) -> None:
        log_connect = True
        while not self.is_closing:
            url = self.connect_url
            if self.reconnect_token is not None:
                url = f"{self.connect_url}/{self.reconnect_token}"
            if log_connect:
                logging.info(f"Connecting To SimplyPrint: {url}")
                log_connect = False
            try:
                self.ws = await tornado.websocket.websocket_connect(
                    url, connect_timeout=5.,
                )
                setattr(self.ws, "on_ping", self._on_ws_ping)
                cur_time = self.eventloop.get_loop_time()
                self._last_ping_received = cur_time
            except asyncio.CancelledError:
                raise
            except Exception:
                curtime = self.eventloop.get_loop_time()
                timediff = curtime - self.last_err_log_time
                if timediff > CONNECTION_ERROR_LOG_TIME:
                    self.last_err_log_time = curtime
                    logging.exception("Failed to connect to SimplyPrint")
            else:
                logging.info("Connected to SimplyPrint Cloud")
                await self._read_messages()
                log_connect = True
            if not self.is_closing:
                await asyncio.sleep(self.reconnect_delay)

    async def _read_messages(self) -> None:
        message: Union[str, bytes, None]
        while self.ws is not None:
            message = await self.ws.read_message()
            if isinstance(message, str):
                self._process_message(message)
            elif message is None:
                self.ping_sp_timer.stop()
                cur_time = self.eventloop.get_loop_time()
                ping_time: float = cur_time - self._last_ping_received
                reason = code = None
                if self.ws is not None:
                    reason = self.ws.close_reason
                    code = self.ws.close_code
                msg = (
                    f"SimplyPrint Disconnected - Code: {code}, "
                    f"Reason: {reason}, "
                    f"Server Ping Time Elapsed: {ping_time}"
                )
                logging.info(msg)
                self.connected = False
                self.ws = None
                break

    def _on_ws_ping(self, data: bytes = b"") -> None:
        self._last_ping_received = self.eventloop.get_loop_time()

    def _process_message(self, msg: str) -> None:
        self._logger.info(f"received: {msg}")
        try:
            packet: Dict[str, Any] = jsonw.loads(msg)
        except jsonw.JSONDecodeError:
            logging.debug(f"Invalid message, not JSON: {msg}")
            return
        event: str = packet.get("type", "")
        data: Optional[Dict[str, Any]] = packet.get("data")
        if event == "connected":
            logging.info("SimplyPrint Reports Connection Success")
            self.connected = True
            self.reconnect_token = None
            if data is not None:
                if data.get("in_setup", 0) == 1:
                    self.is_set_up = False
                    self.save_item("printer_id", None)
                    self.save_item("printer_name", None)
                    if "short_id" in data:
                        self.eventloop.create_task(
                            self._announce_setup(data["short_id"])
                        )
                interval = data.get("interval")
                if isinstance(interval, dict):
                    self._update_intervals(interval)
                self.reconnect_token = data.get("reconnect_token")
                name = data.get("name")
                if name is not None:
                    self.save_item("printer_name", name)
            self.reconnect_delay = 1.
            self._push_initial_state()
            self.ping_sp_timer.start()
        elif event == "error":
            logging.info(f"SimplyPrint Connection Error: {data}")
            self.reconnect_delay = 30.
            self.reconnect_token = None
        elif event == "new_token":
            if data is None:
                logging.debug("Invalid message, no data")
                return
            if data.get("no_exist", False) is True and self.is_set_up:
                self.is_set_up = False
                self.save_item("printer_id", None)
            token: Optional[str] = data.get("token")
            if not isinstance(token, str):
                logging.debug(f"Invalid token received: {token}")
                token = None
            else:
                logging.info("SimplyPrint Token Received")
            self.save_item("printer_token", token)
            self._set_ws_url()
            if "short_id" in data:
                short_id = data["short_id"]
                if not isinstance(short_id, str):
                    self._logger.debug(f"Invalid short_id received: {short_id}")
                else:
                    self.eventloop.create_task(
                        self._announce_setup(data["short_id"])
                    )
        elif event == "complete_setup":
            if data is None:
                logging.debug("Invalid message, no data")
                return
            printer_id = data.get("printer_id")
            if printer_id is None:
                logging.debug("Invalid printer id, received null (None) value")
            self.save_item("printer_id", str(printer_id))
            self._set_ws_url()
            self.save_item("temp_short_setup_id", None)
            self.eventloop.create_task(self._remove_setup_announcement())
        elif event == "demand":
            if data is None:
                logging.debug("Invalid message, no data")
                return
            demand = data.pop("demand", "unknown")
            self._process_demand(demand, data)
        elif event == "interval_change":
            if isinstance(data, dict):
                self._update_intervals(data)
        elif event == "pong":
            diff = self.eventloop.get_loop_time() - self._last_sp_ping
            self.send_sp("latency", {"ms": int(diff * 1000 + .5)})
        else:
            # TODO: It would be good for the backend to send an
            # event indicating that it is ready to recieve printer
            # status.
            logging.debug(f"Unknown event: {msg}")

    def _process_demand(self, demand: str, args: Dict[str, Any]) -> None:
        kconn: KlippyConnection
        kconn = self.server.lookup_component("klippy_connection")
        if demand in ["pause", "resume", "cancel"]:
            if not kconn.is_connected():
                return
            self.eventloop.create_task(self._request_print_action(demand))
        elif demand == "terminal":
            if "enabled" in args:
                self.gcode_terminal_enabled = args["enabled"]
        elif demand == "gcode":
            if not kconn.is_connected():
                return
            script_list: List[str] = args.get("list", [])
            ident: Optional[str] = args.get("identifier", None)
            if script_list:
                script = "\n".join(script_list)
                self.eventloop.create_task(self._handle_gcode_demand(script, ident))
        elif demand == "webcam_snapshot":
            self.eventloop.create_task(self.webcam_stream.post_image(args))
        elif demand == "file":
            url: Optional[str] = args.get("url")
            if not isinstance(url, str):
                logging.debug("Invalid url in message")
                return
            start = bool(args.get("auto_start", 0))
            self.print_handler.download_file(url, start)
        elif demand == "start_print":
            if (
                kconn.is_connected() and
                self.cache.state == "operational"
            ):
                self.eventloop.create_task(self.print_handler.start_print())
            else:
                logging.debug("Failed to start print")
        elif demand == "system_restart":
            coro = self._call_internal_api("machine.reboot")
            self.eventloop.create_task(coro)
        elif demand == "system_shutdown":
            coro = self._call_internal_api("machine.shutdown")
            self.eventloop.create_task(coro)
        elif demand == "api_restart":
            self.eventloop.create_task(self._do_service_action("restart"))
        elif demand == "api_shutdown":
            self.eventloop.create_task(self._do_service_action("shutdown"))
        elif demand == "psu_on":
            self._do_power_action("on")
        elif demand == "psu_off":
            self._do_power_action("off")
        elif demand == "test_webcam":
            self.eventloop.create_task(self._test_webcam())
        else:
            logging.debug(f"Unknown demand: {demand}")

    def save_item(self, name: str, data: Any):
        if data is None:
            self.sp_info.pop(name, None)
            self.spdb.pop(name, None)
        else:
            self.sp_info[name] = data
            self.spdb[name] = data

    async def _handle_gcode_demand(
        self, script: str, ident: Optional[str]
    ) -> None:
        success: bool = True
        msg: Optional[str] = None
        try:
            await self.klippy_apis.run_gcode(script)
        except self.server.error as e:
            msg = str(e)
            success = False
        if ident is not None:
            self.send_sp(
                "gcode_executed",
                {
                    "identifier": ident,
                    "success": success,
                    "message": msg
                }
            )

    async def _call_internal_api(self, method: str, **kwargs) -> Any:
        itransport: InternalTransport
        itransport = self.server.lookup_component("internal_transport")
        try:
            ret = await itransport.call_method(method, **kwargs)
        except self.server.error:
            return None
        return ret

    def _set_ws_url(self) -> None:
        token: Optional[str] = self.sp_info.get("printer_token")
        printer_id: Optional[str] = self.sp_info.get("printer_id")
        ep = TEST_ENDPOINT if self.test else PROD_ENDPOINT
        self.connect_url = f"{ep}/0/0"
        if token is not None:
            if printer_id is None:
                self.connect_url = f"{ep}/0/{token}"
            else:
                self.is_set_up = True
                self.connect_url = f"{ep}/{printer_id}/{token}"

    def _update_intervals(self, intervals: Dict[str, Any]) -> None:
        for key, val in intervals.items():
            self.intervals[key] = val / 1000.
        logging.debug(f"Intervals Updated: {self.intervals}")

    async def _announce_setup(self, short_id: str) -> None:
        async with self.announce_mutex:
            eid: Optional[str] = self.sp_info.get("announcement_id")
            if (
                eid is not None and
                self.sp_info.get("temp_short_setup_id") == short_id
            ):
                return
            ann: Announcements = self.server.lookup_component("announcements")
            if eid is not None:
                # remove stale announcement
                try:
                    await ann.remove_announcement(eid)
                except self.server.error:
                    pass

            self.save_item("temp_short_setup_id", short_id)
            entry = ann.add_internal_announcement(
                "SimplyPrint Setup Request",
                "SimplyPrint is ready to complete setup for your printer. "
                "Please log in to your account and enter the following "
                f"setup code:\n\n{short_id}\n\n",
                "https://simplyprint.io", "high", "simplyprint"
            )
            eid = entry.get("entry_id")
            self.save_item("announcement_id", eid)

    async def _remove_setup_announcement(self) -> None:
        async with self.announce_mutex:
            eid = self.sp_info.get("announcement_id")
            if eid is None:
                return
            self.save_item("announcement_id", None)
            ann: Announcements = self.server.lookup_component("announcements")
            try:
                await ann.remove_announcement(eid)
            except self.server.error:
                pass

    def _do_power_action(self, state: str) -> None:
        if self.power_id:
            power: PrinterPower = self.server.lookup_component("power")
            power.set_device_power(self.power_id, state)

    async def _do_service_action(self, action: str) -> None:
        try:
            machine: Machine = self.server.lookup_component("machine")
            await machine.do_service_action(action, "moonraker")
        except self.server.error:
            pass

    async def _request_print_action(self, action: str) -> None:
        cur_state = self.cache.state
        ret: Optional[str] = ""
        self._print_request_event.clear()
        if action == "pause":
            if cur_state == "printing":
                self._update_state("pausing")
                ret = await self.klippy_apis.pause_print(None)
        elif action == "resume":
            if cur_state == "paused":
                self._print_request_fut = self.eventloop.create_future()
                self._update_state("resuming")
                ret = await self.klippy_apis.resume_print(None)
        elif action == "cancel":
            if cur_state in ["printing", "paused"]:
                self._update_state("cancelling")
                ret = await self.klippy_apis.cancel_print(None)
        if ret is None:
            # Wait for the "action" requested event to fire, then reset the
            # state
            try:
                await asyncio.wait_for(self._print_request_event.wait(), 1.)
            except Exception:
                pass
            self._update_state_from_klippy()

    async def _test_webcam(self) -> None:
        await self.webcam_stream.test_connection()
        self.send_sp(
            "webcam_status", {"connected": self.webcam_stream.connected}
        )

    async def _on_klippy_ready(self) -> None:
        last_stats: Dict[str, Any] = self.job_state.get_last_stats()
        if last_stats["state"] == "printing":
            self._on_print_started(last_stats, last_stats, False)
        else:
            self._update_state("operational")
        query: Optional[Dict[str, Any]]
        query = await self.klippy_apis.query_objects({"heaters": None}, None)
        sub_objs = {
            "display_status": ["progress"],
            "bed_mesh": ["mesh_matrix", "mesh_min", "mesh_max"],
            "toolhead": ["extruder"],
            "gcode_move": ["gcode_position"]
        }
        # Add Heater Subscriptions
        has_amb_sensor: bool = False
        cfg_amb_sensor = self.amb_detect.sensor_name
        if query is not None:
            heaters: Dict[str, Any] = query.get("heaters", {})
            avail_htrs: List[str]
            avail_htrs = sorted(heaters.get("available_heaters", []))
            logging.debug(f"SimplyPrint: Heaters Detected: {avail_htrs}")
            for htr in avail_htrs:
                if htr.startswith("extruder"):
                    sub_objs[htr] = ["temperature", "target"]
                    if htr == "extruder":
                        tool_id = "tool0"
                    else:
                        tool_id = "tool" + htr[8:]
                    self.heaters[htr] = tool_id
                elif htr == "heater_bed":
                    sub_objs[htr] = ["temperature", "target"]
                    self.heaters[htr] = "bed"
            sensors: List[str] = heaters.get("available_sensors", [])
            if cfg_amb_sensor:
                if cfg_amb_sensor in sensors:
                    has_amb_sensor = True
                    sub_objs[cfg_amb_sensor] = ["temperature"]
                else:
                    logging.info(
                        f"SimplyPrint: Ambient sensor {cfg_amb_sensor} not "
                        "configured in Klipper"
                    )
        if self.filament_sensor:
            objects: List[str]
            objects = await self.klippy_apis.get_object_list(default=[])
            if self.filament_sensor in objects:
                sub_objs[self.filament_sensor] = ["filament_detected"]
            # Add filament sensor subscription
        if not sub_objs:
            return
        # Create our own subscription rather than use the host sub
        status: Dict[str, Any] = await self.klippy_apis.subscribe_from_transport(
            sub_objs, self, default={}
        )
        if status:
            logging.debug(f"SimplyPrint: Got Initial Status: {status}")
            self.printer_status = status
            self._update_temps(1.)
            self.next_temp_update_time = 0.
            if "bed_mesh" in status:
                self._send_mesh_data()
            if "toolhead" in status:
                self._send_active_extruder(status["toolhead"]["extruder"])
            if "gcode_move" in status:
                self.layer_detect.update(
                    status["gcode_move"]["gcode_position"]
                )
            if self.filament_sensor and self.filament_sensor in status:
                detected = status[self.filament_sensor]["filament_detected"]
                fstate = "loaded" if detected else "runout"
                self.cache.filament_state = fstate
                self.send_sp("filament_sensor", {"state": fstate})
            if has_amb_sensor and cfg_amb_sensor in status:
                self.amb_detect.update_ambient(status[cfg_amb_sensor])
        if not has_amb_sensor:
            self.amb_detect.start()
        self.printer_info_timer.start(delay=1.)

    def _on_power_changed(self, device_info: Dict[str, Any]) -> None:
        if self.power_id and device_info["device"] == self.power_id:
            is_on = device_info["status"] == "on"
            self.send_sp("power_controller", {"on": is_on})

    def _on_websocket_identified(self, ws: BaseRemoteConnection) -> None:
        if (
            self.cache.current_wsid is None and
            ws.client_data.get("type", "") == "web"
        ):
            ui_data: Dict[str, Any] = {
                "ui": ws.client_data["name"],
                "ui_version": ws.client_data["version"]
            }
            self.cache.firmware_info.update(ui_data)
            self.cache.current_wsid = ws.uid
            self.send_sp("machine_data", ui_data)

    def _on_websocket_removed(self, ws: BaseRemoteConnection) -> None:
        if self.cache.current_wsid is None or self.cache.current_wsid != ws.uid:
            return
        ui_data = self._get_ui_info()
        diff = self._get_object_diff(ui_data, self.cache.firmware_info)
        if diff:
            self.cache.firmware_info.update(ui_data)
            self.send_sp("machine_data", ui_data)

    def _on_klippy_startup(self, state: KlippyState) -> None:
        if state != KlippyState.READY:
            self._update_state("error")
            kconn: KlippyConnection
            kconn = self.server.lookup_component("klippy_connection")
            self.send_sp("printer_error", {"error": kconn.state.message})
        self.send_sp("connection", {"new": "connected"})
        self._send_firmware_data()

    def _on_klippy_shutdown(self) -> None:
        self._update_state("error")
        kconn: KlippyConnection
        kconn = self.server.lookup_component("klippy_connection")
        self.send_sp("printer_error", {"error": kconn.state.message})

    def _on_klippy_disconnected(self) -> None:
        self._update_state("offline")
        self.send_sp("connection", {"new": "disconnected"})
        self.amb_detect.stop()
        self.printer_info_timer.stop()
        self.cache.reset_print_state()
        self.printer_status = {}

    def _on_job_state_changed(self, job_event: JobEvent, *args) -> None:
        callback: Optional[Callable] = getattr(self, f"_on_print_{job_event}", None)
        if callback is not None:
            callback(*args)
        else:
            logging.info(f"No defined callback for Job Event: {job_event}")

    def _on_print_started(
        self,
        prev_stats: Dict[str, Any],
        new_stats: Dict[str, Any],
        need_start_event: bool = True
    ) -> None:
        # inlcludes started and resumed events
        self._update_state("printing")
        filename = new_stats["filename"]
        job_info: Dict[str, Any] = {"filename": filename}
        fm: FileManager = self.server.lookup_component("file_manager")
        metadata = fm.get_file_metadata(filename)
        filament: Optional[float] = metadata.get("filament_total")
        if filament is not None:
            job_info["filament"] = round(filament)
        est_time = metadata.get("estimated_time")
        if est_time is not None:
            job_info["time"] = est_time
        self.cache.metadata = metadata
        self.cache.job_info.update(job_info)
        if need_start_event:
            job_info["started"] = True
        self.layer_detect.start(metadata)
        self._send_job_event(job_info)

    def _check_job_started(
        self,
        prev_stats: Dict[str, Any],
        new_stats: Dict[str, Any]
    ) -> None:
        if not self.cache.job_info:
            job_info: Dict[str, Any] = {
                "filename": new_stats.get("filename", ""),
                "started": True
            }
            self._send_job_event(job_info)

    def _reset_file(self) -> None:
        cur_job = self.cache.job_info.get("filename", "")
        last_started = self.print_handler.last_started
        if last_started and last_started == cur_job:
            kapi: KlippyAPI = self.server.lookup_component("klippy_apis")
            self.eventloop.create_task(
                kapi.run_gcode("SDCARD_RESET_FILE", default=None)
            )
        self.print_handler.last_started = ""

    def _on_print_paused(self, *args) -> None:
        self.send_sp("job_info", {"paused": True})
        self._update_state("paused")
        self.layer_detect.stop()

    def _on_print_resumed(self, *args) -> None:
        self._update_state("printing")
        self.layer_detect.resume()

    def _on_print_cancelled(self, *args) -> None:
        self._check_job_started(*args)
        self._reset_file()
        self._send_job_event({"cancelled": True})
        self._update_state_from_klippy()
        self.cache.job_info = {}
        self.layer_detect.stop()

    def _on_print_error(self, *args) -> None:
        self._check_job_started(*args)
        self._reset_file()
        payload: Dict[str, Any] = {"failed": True}
        new_stats: Dict[str, Any] = args[1]
        msg = new_stats.get("message", "Unknown Error")
        payload["error"] = msg
        self._send_job_event(payload)
        self._update_state_from_klippy()
        self.cache.job_info = {}
        self.layer_detect.stop()

    def _on_print_complete(self, *args) -> None:
        self._check_job_started(*args)
        self._reset_file()
        self._send_job_event({"finished": True})
        self._update_state_from_klippy()
        self.cache.job_info = {}
        self.layer_detect.stop()

    def _on_print_standby(self, *args) -> None:
        self._update_state_from_klippy()
        self.cache.job_info = {}
        self.layer_detect.stop()

    def _on_pause_requested(self) -> None:
        self._print_request_event.set()
        if self.cache.state == "printing":
            self._update_state("pausing")

    def _on_resume_requested(self) -> None:
        self._print_request_event.set()
        if self.cache.state == "paused":
            self._update_state("resuming")

    def _on_cancel_requested(self) -> None:
        self._print_request_event.set()
        if self.cache.state in ["printing", "paused", "pausing"]:
            self._update_state("cancelling")

    def _on_gcode_response(self, response: str):
        if self.gcode_terminal_enabled:
            resp = [
                r.strip() for r in response.strip().split("\n") if r.strip()
            ]
            self.send_sp("term_update", {"response": resp})

    def _on_gcode_received(self, script: str):
        if self.gcode_terminal_enabled:
            cmds = [s.strip() for s in script.strip().split() if s.strip()]
            self.send_sp("term_update", {"command": cmds})

    def _on_proc_update(self, proc_stats: Dict[str, Any]) -> None:
        cpu = proc_stats["system_cpu_usage"]
        if not cpu:
            return
        curtime = self.eventloop.get_loop_time()
        if curtime - self.last_cpu_update_time < self.intervals["cpu"]:
            return
        self.last_cpu_update_time = curtime
        sys_mem = proc_stats["system_memory"]
        mem_pct: float = 0.
        if sys_mem:
            mem_pct = sys_mem["used"] / sys_mem["total"] * 100
        cpu_data = {
            "usage": int(cpu["cpu"] + .5),
            "memory": int(mem_pct + .5),
            "flags": self.cache.throttled_state.get("bits", 0)
        }
        temp: Optional[float] = proc_stats["cpu_temp"]
        if temp is not None:
            cpu_data["temp"] = int(temp + .5)
        diff = self._get_object_diff(cpu_data, self.cache.cpu_info)
        if diff:
            self.cache.cpu_info.update(cpu_data)
            self.send_sp("cpu", diff)

    def _on_cpu_throttled(self, throttled_state: Dict[str, Any]):
        self.cache.throttled_state = throttled_state

    def send_status(self, status: Dict[str, Any], eventtime: float) -> None:
        for printer_obj, vals in status.items():
            self.printer_status[printer_obj].update(vals)
        if self.amb_detect.sensor_name in status:
            self.amb_detect.update_ambient(
                status[self.amb_detect.sensor_name], eventtime
            )
        self._update_temps(eventtime)
        if "bed_mesh" in status:
            self._send_mesh_data()
        if "toolhead" in status and "extruder" in status["toolhead"]:
            self._send_active_extruder(status["toolhead"]["extruder"])
        if "gcode_move" in status:
            self.layer_detect.update(status["gcode_move"]["gcode_position"])
        if self.filament_sensor and self.filament_sensor in status:
            detected = status[self.filament_sensor]["filament_detected"]
            fstate = "loaded" if detected else "runout"
            self.cache.filament_state = fstate
            self.send_sp("filament_sensor", {"state": fstate})

    def _handle_printer_info_update(self, eventtime: float) -> float:
        # Job Info Timer handler
        if self.cache.state == "printing":
            self._update_job_progress()
        return eventtime + self.intervals["job"]

    def _handle_sp_ping(self, eventtime: float) -> float:
        self._last_sp_ping = eventtime
        self.send_sp("ping", None)
        return eventtime + self.intervals["ping"]

    def _update_job_progress(self) -> None:
        job_info: Dict[str, Any] = {}
        est_time = self.cache.metadata.get("estimated_time")
        last_stats: Dict[str, Any] = self.job_state.get_last_stats()
        if est_time is not None:
            duration: float = last_stats["print_duration"]
            time_left = max(0, int(est_time - duration + .5))
            last_time_left = self.cache.job_info.get("time", time_left + 60.)
            time_diff = last_time_left - time_left
            if (
                (time_left < 60 or time_diff >= 30) and
                time_left != last_time_left
            ):
                job_info["time"] = time_left
        if "display_status" in self.printer_status:
            progress = self.printer_status["display_status"]["progress"]
            pct_prog = int(progress * 100 + .5)
            if pct_prog != self.cache.job_info.get("progress", 0):
                job_info["progress"] = int(progress * 100 + .5)
        layer: Optional[int] = last_stats.get("info", {}).get("current_layer")
        if layer is None:
            layer = self.layer_detect.layer
        if layer != self.cache.job_info.get("layer", -1):
            job_info["layer"] = layer
        if job_info:
            self.cache.job_info.update(job_info)
            self.send_sp("job_info", job_info)

    def _update_temps(self, eventtime: float) -> None:
        if eventtime < self.next_temp_update_time:
            return
        need_rapid_update: bool = False
        temp_data: Dict[str, List[int]] = {}
        for printer_obj, key in self.heaters.items():
            reported_temp = self.printer_status[printer_obj]["temperature"]
            ret = [
                int(reported_temp + .5),
                int(self.printer_status[printer_obj]["target"] + .5)
            ]
            last_temps = self.cache.temps.get(key, [-100., -100.])
            if ret[1] == last_temps[1]:
                if ret[1]:
                    seeking_target = abs(ret[1] - ret[0]) > 5
                else:
                    seeking_target = ret[0] >= self.amb_detect.ambient + 25
                need_rapid_update |= seeking_target
                # The target hasn't changed and not heating, debounce temp
                if key in self.last_received_temps and not seeking_target:
                    last_reported = self.last_received_temps[key]
                    if abs(reported_temp - last_reported) < .75:
                        self.last_received_temps.pop(key)
                        continue
                if ret[0] == last_temps[0]:
                    self.last_received_temps[key] = reported_temp
                    continue
                temp_data[key] = ret[:1]
            else:
                # target has changed, send full data
                temp_data[key] = ret
            self.last_received_temps[key] = reported_temp
            self.cache.temps[key] = ret
        if need_rapid_update:
            self.next_temp_update_time = (
                0. if self.intervals["temps_target"] < .2501 else
                eventtime + self.intervals["temps_target"]
            )
        else:
            self.next_temp_update_time = eventtime + self.intervals["temps"]
        if not temp_data:
            return
        self.send_sp("temps", temp_data)

    def _update_state_from_klippy(self) -> None:
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        klippy_state = kconn.state
        if klippy_state == KlippyState.READY:
            sp_state = "operational"
        elif klippy_state in [KlippyState.ERROR, KlippyState.SHUTDOWN]:
            sp_state = "error"
        else:
            sp_state = "offline"
        self._update_state(sp_state)

    def _update_state(self, new_state: str) -> None:
        if self.cache.state == new_state:
            return
        self.cache.state = new_state
        self.send_sp("state_change", {"new": new_state})
        if new_state == "operational":
            self.print_handler.notify_ready()

    def _send_mesh_data(self) -> None:
        mesh = self.printer_status["bed_mesh"]
        # TODO: We are probably going to have to reformat the mesh
        self.cache.mesh = mesh
        self.send_sp("mesh_data", mesh)

    def _send_job_event(self, job_info: Dict[str, Any]) -> None:
        if self.connected:
            self.send_sp("job_info", job_info)
        else:
            job_info.update(self.cache.job_info)
            job_info["delay"] = self.eventloop.get_loop_time()
            self.missed_job_events.append(job_info)
            if len(self.missed_job_events) > 10:
                self.missed_job_events.pop(0)

    def _get_ui_info(self) -> Dict[str, Any]:
        ui_data: Dict[str, Any] = {"ui": None, "ui_version": None}
        self.cache.current_wsid = None
        websockets: WebsocketManager
        websockets = self.server.lookup_component("websockets")
        conns = websockets.get_clients_by_type("web")
        if conns:
            longest = conns[0]
            ui_data["ui"] = longest.client_data["name"]
            ui_data["ui_version"] = longest.client_data["version"]
            self.cache.current_wsid = longest.uid
        return ui_data

    async def _send_machine_data(self) -> None:
        app_args = self.server.get_app_args()
        data = self._get_ui_info()
        data["api"] = "Moonraker"
        data["api_version"] = app_args["software_version"]
        data["sp_version"] = COMPONENT_VERSION
        machine: Machine = self.server.lookup_component("machine")
        sys_info = machine.get_system_info()
        pyver = sys_info["python"]["version"][:3]
        data["python_version"] = ".".join([str(part) for part in pyver])
        model: str = sys_info["cpu_info"].get("model", "")
        if not model or model.isdigit():
            model = sys_info["cpu_info"].get("cpu_desc", "Unknown")
        data["machine"] = model
        data["os"] = sys_info["distribution"].get("name", "Unknown")
        pub_intf = await machine.get_public_network()
        data["is_ethernet"] = int(not pub_intf["is_wifi"])
        data["ssid"] = pub_intf.get("ssid", "")
        data["local_ip"] = pub_intf.get("address", "Unknown")
        data["hostname"] = pub_intf["hostname"]
        data["core_count"] = os.cpu_count()
        mem = sys_info["cpu_info"]["total_memory"]
        if mem is not None:
            data["total_memory"] = mem * 1024
        self.cache.machine_info = data
        self.send_sp("machine_data", data)

    def _send_firmware_data(self) -> None:
        kinfo = self.server.get_klippy_info()
        if "software_version" not in kinfo:
            return
        firmware_date: str = ""
        # Approximate the firmware "date" using the last modified
        # time of the Klippy source folder
        kpath = pathlib.Path(kinfo["klipper_path"]).joinpath("klippy")
        if kpath.is_dir():
            mtime = kpath.stat().st_mtime
            firmware_date = time.asctime(time.gmtime(mtime))
        version: str = kinfo["software_version"]
        unsafe = version.endswith("-dirty") or version == "?"
        if unsafe:
            version = version.rsplit("-", 1)[0]
        fw_info = {
            "firmware": "Klipper",
            "firmware_version": version,
            "firmware_date": firmware_date,
            "firmware_link": "https://github.com/Klipper3d/klipper",
        }
        diff = self._get_object_diff(fw_info, self.cache.firmware_info)
        if diff:
            self.cache.firmware_info = fw_info
            self.send_sp(
                "firmware", {"fw": diff, "raw": False, "unsafe": unsafe}
            )

    def _send_active_extruder(self, new_extruder: str):
        tool = "T0" if new_extruder == "extruder" else f"T{new_extruder[8:]}"
        if tool == self.cache.active_extruder:
            return
        self.cache.active_extruder = tool
        self.send_sp("tool", {"new": tool})

    async def _send_webcam_config(self) -> None:
        wc_cfg = await self.webcam_stream.get_webcam_config()
        wc_data = {
            "flipH": wc_cfg.get("flip_horizontal", False),
            "flipV": wc_cfg.get("flip_vertical", False),
            "rotate90": wc_cfg.get("rotation", 0) == 90
        }
        self.send_sp("webcam", wc_data)

    async def _send_power_state(self) -> None:
        dev_info: Optional[Dict[str, Any]]
        dev_info = await self._call_internal_api(
            "machine.device_power.get_device", device=self.power_id
        )
        if dev_info is not None:
            is_on = dev_info[self.power_id] == "on"
            self.send_sp("power_controller", {"on": is_on})

    def _push_initial_state(self):
        self.send_sp("state_change", {"new": self.cache.state})
        if self.cache.temps:
            self.send_sp("temps", self.cache.temps)
        if self.cache.firmware_info:
            self.send_sp(
                "firmware",
                {"fw": self.cache.firmware_info, "raw": False})
        curtime = self.eventloop.get_loop_time()
        for evt in self.missed_job_events:
            evt["delay"] = int((curtime - evt["delay"]) + .5)
            self.send_sp("job_info", evt)
        self.missed_job_events = []
        if self.cache.active_extruder:
            self.send_sp("tool", {"new": self.cache.active_extruder})
        if self.cache.cpu_info:
            self.send_sp("cpu_info", self.cache.cpu_info)
        self.send_sp("ambient", {"new": self.amb_detect.ambient})
        if self.power_id:
            self.eventloop.create_task(self._send_power_state())
        if self.cache.filament_state:
            self.send_sp(
                "filament_sensor", {"state": self.cache.filament_state}
            )
        self.send_sp(
            "webcam_status", {"connected": self.webcam_stream.connected}
        )
        self.eventloop.create_task(self._send_machine_data())
        self.eventloop.create_task(self._send_webcam_config())

    def _check_setup_event(self, evt_name: str) -> bool:
        return self.is_set_up or evt_name in PRE_SETUP_EVENTS

    def send_sp(self, evt_name: str, data: Any) -> Awaitable[bool]:
        if (
            not self.connected or
            self.ws is None or
            self.ws.protocol is None or
            not self._check_setup_event(evt_name)
        ):
            fut = self.eventloop.create_future()
            fut.set_result(False)
            return fut
        packet = {"type": evt_name, "data": data}
        return self.eventloop.create_task(self._send_wrapper(packet))

    async def _send_wrapper(self, packet: Dict[str, Any]) -> bool:
        try:
            assert self.ws is not None
            await self.ws.write_message(jsonw.dumps(packet))
        except Exception:
            return False
        else:
            if packet["type"] != "stream":
                self._logger.info(f"sent: {packet}")
            else:
                self._logger.info("sent: webcam stream")
        return True

    def _get_object_diff(
        self, new_obj: Dict[str, Any], cached_obj: Dict[str, Any]
    ) -> Dict[str, Any]:
        if not cached_obj:
            return new_obj
        diff: Dict[str, Any] = {}
        for key, val in new_obj.items():
            if key in cached_obj and val == cached_obj[key]:
                continue
            diff[key] = val
        return diff

    async def close(self):
        self.print_handler.cancel()
        self.amb_detect.stop()
        self.printer_info_timer.stop()
        self.ping_sp_timer.stop()
        await self.send_sp("shutdown", None)
        self._logger.close()
        self.is_closing = True
        if self.ws is not None:
            self.ws.close(1001, "Client Shutdown")
        if (
            self.connection_task is not None and
            not self.connection_task.done()
        ):
            try:
                await asyncio.wait_for(self.connection_task, 2.)
            except asyncio.TimeoutError:
                pass

class ReportCache:
    def __init__(self) -> None:
        self.state = "offline"
        self.temps: Dict[str, Any] = {}
        self.metadata: Dict[str, Any] = {}
        self.mesh: Dict[str, Any] = {}
        self.job_info: Dict[str, Any] = {}
        self.active_extruder: str = ""
        # Persistent state across connections
        self.firmware_info: Dict[str, Any] = {}
        self.machine_info: Dict[str, Any] = {}
        self.cpu_info: Dict[str, Any] = {}
        self.throttled_state: Dict[str, Any] = {}
        self.current_wsid: Optional[int] = None
        self.filament_state: str = ""

    def reset_print_state(self) -> None:
        self.temps = {}
        self.mesh = {}
        self.job_info = {}


INITIAL_AMBIENT = 85
AMBIENT_CHECK_TIME = 5. * 60.
TARGET_CHECK_TIME = 60. * 60.
SAMPLE_CHECK_TIME = 20.

class AmbientDetect:
    CHECK_INTERVAL = 5
    def __init__(
        self,
        config: ConfigHelper,
        simplyprint: SimplyPrint,
        initial_ambient: int
    ) -> None:
        self.server = config.get_server()
        self.simplyprint = simplyprint
        self.cache = simplyprint.cache
        self._initial_sample: int = -1000
        self._ambient = initial_ambient
        self._last_sample_time: float = 0.
        self._update_interval = AMBIENT_CHECK_TIME
        self.eventloop = self.server.get_event_loop()
        self._detect_timer = self.eventloop.register_timer(
            self._handle_detect_timer
        )
        self._sensor_name: str = config.get("ambient_sensor", "")

    @property
    def ambient(self) -> int:
        return self._ambient

    @property
    def sensor_name(self) -> str:
        return self._sensor_name

    def update_ambient(
        self, sensor_info: Dict[str, Any], eventtime: float = SAMPLE_CHECK_TIME
    ) -> None:
        if "temperature" not in sensor_info:
            return
        if eventtime < self._last_sample_time + SAMPLE_CHECK_TIME:
            return
        self._last_sample_time = eventtime
        new_amb = int(sensor_info["temperature"] + .5)
        if abs(new_amb - self._ambient) < 2:
            return
        self._ambient = new_amb
        self._on_ambient_changed(self._ambient)

    def _handle_detect_timer(self, eventtime: float) -> float:
        if "tool0" not in self.cache.temps:
            self._initial_sample = -1000
            return eventtime + self.CHECK_INTERVAL
        temp, target = self.cache.temps["tool0"]
        if target:
            self._initial_sample = -1000
            self._last_sample_time = eventtime
            self._update_interval = TARGET_CHECK_TIME
            return eventtime + self.CHECK_INTERVAL
        if eventtime - self._last_sample_time < self._update_interval:
            return eventtime + self.CHECK_INTERVAL
        if self._initial_sample == -1000:
            self._initial_sample = temp
            self._update_interval = SAMPLE_CHECK_TIME
        else:
            diff = abs(temp - self._initial_sample)
            if diff <= 2:
                last_ambient = self._ambient
                self._ambient = int((temp + self._initial_sample) / 2 + .5)
                self._initial_sample = -1000
                self._last_sample_time = eventtime
                self._update_interval = AMBIENT_CHECK_TIME
                if last_ambient != self._ambient:
                    logging.debug(f"SimplyPrint: New Ambient: {self._ambient}")
                    self._on_ambient_changed(self._ambient)
            else:
                self._initial_sample = temp
                self._update_interval = SAMPLE_CHECK_TIME
        return eventtime + self.CHECK_INTERVAL

    def _on_ambient_changed(self, new_ambient: int) -> None:
        self.simplyprint.save_item("ambient_temp", new_ambient)
        self.simplyprint.send_sp("ambient", {"new": new_ambient})

    def start(self) -> None:
        if self._detect_timer.is_running():
            return
        if "tool0" in self.cache.temps:
            cur_temp = self.cache.temps["tool0"][0]
            if cur_temp < self._ambient:
                self._ambient = cur_temp
                self._on_ambient_changed(self._ambient)
        self._detect_timer.start()

    def stop(self) -> None:
        self._detect_timer.stop()
        self._last_sample_time = 0.

class LayerDetect:
    def __init__(self) -> None:
        self._layer: int = 0
        self._layer_z: float = 0.
        self._active: bool = False
        self._layer_height: float = 0.
        self._fl_height: float = 0.
        self._layer_count: int = 99999999999
        self._check_next: bool = False

    @property
    def layer(self) -> int:
        return self._layer

    def update(self, new_pos: List[float]) -> None:
        if not self._active or self._layer_z == new_pos[2]:
            self._check_next = False
            return
        if not self._check_next:
            # Try to avoid z-hops by skipping the first detected change
            self._check_next = True
            return
        self._check_next = False
        layer = 1 + int(
            (new_pos[2] - self._fl_height) / self._layer_height + .5
        )
        self._layer = min(layer, self._layer_count)
        self._layer_z = new_pos[2]

    def start(self, metadata: Dict[str, Any]) -> None:
        self.reset()
        lh: float = metadata.get("layer_height", 0)
        flh: float = metadata.get("first_layer_height", lh)
        if lh > 0.000001 and flh > 0.000001:
            self._active = True
            self._layer_height = lh
            self._fl_height = flh
            layer_count: Optional[int] = metadata.get("layer_count")
            obj_height: Optional[float] = metadata.get("object_height")
            if layer_count is not None:
                self._layer_count = layer_count
            elif obj_height is not None:
                self._layer_count = int((obj_height - flh) / lh + .5)

    def resume(self) -> None:
        self._active = True

    def stop(self) -> None:
        self._active = False

    def reset(self) -> None:
        self._active = False
        self._layer = 0
        self._layer_z = 0.
        self._layer_height = 0.
        self._fl_height = 0.
        self._layer_count = 99999999999
        self._check_next = False


# TODO: We need to get the URL/Port from settings in the future.
# Ideally we will always fetch from the localhost rather than
# go through the reverse proxy
FALLBACK_URL = "http://127.0.0.1:8080/?action=snapshot"
SP_SNAPSHOT_URL = "https://api.simplyprint.io/jobs/ReceiveSnapshot"

class WebcamStream:
    def __init__(
        self, config: ConfigHelper, simplyprint: SimplyPrint
    ) -> None:
        self.server = config.get_server()
        self.eventloop = self.server.get_event_loop()
        self.simplyprint = simplyprint
        self.webcam_name = config.get("webcam_name", "")
        self.url = FALLBACK_URL
        self.client: HttpClient = self.server.lookup_component("http_client")
        self.cam: Optional[WebCam] = None
        self._connected = False

    @property
    def connected(self) -> bool:
        return self._connected

    async def intialize_url(self) -> None:
        wcmgr: WebcamManager = self.server.lookup_component("webcam")
        cams = wcmgr.get_webcams()
        if not cams:
            # no camera configured, try the fallback url
            return
        if self.webcam_name and self.webcam_name in cams:
            cam = cams[self.webcam_name]
        else:
            cam = list(cams.values())[0]
        try:
            url = await cam.get_snapshot_url(True)
        except Exception:
            logging.exception("Failed to retrive webcam url")
            return
        self.cam = cam
        logging.info(f"SimplyPrint Webcam URL assigned: {url}")
        self.url = url

    async def test_connection(self):
        if not self.url.startswith("http"):
            self._connected = False
            return
        headers = {"Accept": "image/jpeg"}
        resp = await self.client.get(self.url, headers, enable_cache=False)
        self._connected = not resp.has_error()

    async def get_webcam_config(self) -> Dict[str, Any]:
        if self.cam is None:
            return {}
        return self.cam.as_dict()

    async def extract_image(self) -> str:
        headers = {"Accept": "image/jpeg"}
        resp = await self.client.get(self.url, headers, enable_cache=False)
        resp.raise_for_status()
        return await self.eventloop.run_in_thread(
            self._encode_image, resp.content
        )

    def _encode_image(self, image: bytes) -> str:
        return base64.b64encode(image).decode()

    async def post_image(self, payload: Dict[str, Any]) -> None:
        uid: Optional[str] = payload.get("id")
        timer: Optional[int] = payload.get("timer")
        try:
            if uid is not None:
                url = payload.get("endpoint", SP_SNAPSHOT_URL)
                img = await self.extract_image()
                headers = {
                    "User-Agent": "Mozilla/5.0",
                    "Content-Type": "application/x-www-form-urlencoded"
                }
                body = f"id={url_escape(uid)}&image={url_escape(img)}"
                resp = await self.client.post(
                    url, body=body, headers=headers, enable_cache=False
                )
                resp.raise_for_status()
            elif timer is not None:
                await asyncio.sleep(timer / 1000)
                img = await self.extract_image()
                self.simplyprint.send_sp("stream", {"base": img})
        except asyncio.CancelledError:
            raise
        except Exception:
            if not self.server.is_verbose_enabled():
                return
            logging.exception("SimplyPrint WebCam Stream Error")

class PrintHandler:
    def __init__(self, simplyprint: SimplyPrint) -> None:
        self.simplyprint = simplyprint
        self.server = simplyprint.server
        self.eventloop = self.server.get_event_loop()
        self.cache = simplyprint.cache
        self.download_task: Optional[asyncio.Task] = None
        self.print_ready_event: asyncio.Event = asyncio.Event()
        self.download_progress: int = -1
        self.pending_file: str = ""
        self.last_started: str = ""
        self.sp_user = UserInfo("SimplyPrint", "")

    def download_file(self, url: str, start: bool):
        coro = self._download_sp_file(url, start)
        self.download_task = self.eventloop.create_task(coro)

    def cancel(self):
        if (
            self.download_task is not None and
            not self.download_task.done()
        ):
            self.download_task.cancel()
            self.download_task = None

    def notify_ready(self):
        self.print_ready_event.set()

    async def _download_sp_file(self, url: str, start: bool):
        client: HttpClient = self.server.lookup_component("http_client")
        fm: FileManager = self.server.lookup_component("file_manager")
        gc_path = pathlib.Path(fm.get_directory())
        if not gc_path.is_dir():
            logging.debug(f"GCode Path Not Registered: {gc_path}")
            self.simplyprint.send_sp(
                "file_progress",
                {"state": "error", "message": "GCode Path not Registered"}
            )
            return
        accept = "text/plain,applicaton/octet-stream"
        self._on_download_progress(0, 0, 0)
        try:
            logging.debug(f"Downloading URL: {url}")
            tmp_path = await client.download_file(
                url, accept, progress_callback=self._on_download_progress,
                request_timeout=3600.
            )
        except asyncio.TimeoutError:
            raise
        except Exception:
            logging.exception(f"Failed to download file: {url}")
            self.simplyprint.send_sp(
                "file_progress",
                {"state": "error", "message": "Network Error"}
            )
            return
        finally:
            self.download_progress = -1
        logging.debug("Simplyprint: Download Complete")
        filename = pathlib.PurePath(tmp_path.name)
        fpath = gc_path.joinpath(filename.name)
        if self.cache.job_info.get("filename", "") == str(fpath):
            # This is an attempt to overwite a print in progress, make a copy
            count = 0
            while fpath.exists():
                name = f"{filename.stem}_copy_{count}.{filename.suffix}"
                fpath = gc_path.joinpath(name)
                count += 1
        args: Dict[str, Any] = {
            "filename": fpath.name,
            "tmp_file_path": str(tmp_path),
        }
        state = "pending"
        if self.cache.state == "operational":
            args["print"] = "true" if start else "false"
        try:
            ret = await fm.finalize_upload(args)
        except self.server.error as e:
            logging.exception("GCode Finalization Failed")
            self.simplyprint.send_sp(
                "file_progress",
                {"state": "error", "message": f"GCode Finalization Failed: {e}"}
            )
            return
        self.pending_file = fpath.name
        if ret.get("print_started", False):
            state = "started"
            self.last_started = self.pending_file
            self.pending_file = ""
        elif not start and await self._check_can_print():
            state = "ready"
        if state == "pending":
            self.print_ready_event.clear()
            try:
                await asyncio.wait_for(self.print_ready_event.wait(), 10.)
            except asyncio.TimeoutError:
                self.pending_file = ""
                self.simplyprint.send_sp(
                    "file_progress",
                    {"state": "error", "message": "Pending print timed out"}
                )
                return
            else:
                if start:
                    await self.start_print()
                    return
                state = "ready"
        self.simplyprint.send_sp("file_progress", {"state": state})

    async def start_print(self) -> None:
        if not self.pending_file:
            return
        pending = self.pending_file
        self.pending_file = ""
        kapi: KlippyAPI = self.server.lookup_component("klippy_apis")
        data = {"state": "started"}
        try:
            await kapi.start_print(pending, user=self.sp_user)
        except Exception:
            logging.exception("Print Failed to start")
            data["state"] = "error"
            data["message"] = "Failed to start print"
        else:
            self.last_started = pending
        self.simplyprint.send_sp("file_progress", data)

    async def _check_can_print(self) -> bool:
        kconn: KlippyConnection = self.server.lookup_component("klippy_connection")
        if kconn.state != KlippyState.READY:
            return False
        kapi: KlippyAPI = self.server.lookup_component("klippy_apis")
        try:
            result = await kapi.query_objects({"print_stats": None})
        except Exception:
            # Klippy not connected
            return False
        if 'print_stats' not in result:
            return False
        state: str = result['print_stats']['state']
        if state in ["printing", "paused"]:
            return False
        return True

    def _on_download_progress(self, percent: int, size: int, recd: int) -> None:
        if percent == self.download_progress:
            return
        self.download_progress = percent
        self.simplyprint.send_sp(
            "file_progress", {"state": "downloading", "percent": percent}
        )

class ProtoLogger:
    def __init__(self, config: ConfigHelper) -> None:
        server = config.get_server()
        self._logger: Optional[logging.Logger] = None
        if not server.is_verbose_enabled():
            return
        fm: FileManager = server.lookup_component("file_manager")
        log_root = fm.get_directory("logs")
        if log_root:
            log_parent = pathlib.Path(log_root)
        else:
            log_parent = pathlib.Path(tempfile.gettempdir())
        log_path = log_parent.joinpath("simplyprint.log")
        queue: SimpleQueue = SimpleQueue()
        self.queue_handler = LocalQueueHandler(queue)
        self._logger = logging.getLogger("simplyprint")
        self._logger.addHandler(self.queue_handler)
        self._logger.propagate = False
        file_hdlr = logging.handlers.TimedRotatingFileHandler(
            log_path, when='midnight', backupCount=2)
        formatter = logging.Formatter(
            '%(asctime)s [%(funcName)s()] - %(message)s')
        file_hdlr.setFormatter(formatter)
        self.qlistner = logging.handlers.QueueListener(queue, file_hdlr)
        self.qlistner.start()

    def info(self, msg: str) -> None:
        if self._logger is None:
            return
        self._logger.info(msg)

    def debug(self, msg: str) -> None:
        if self._logger is None:
            return
        self._logger.debug(msg)

    def warning(self, msg: str) -> None:
        if self._logger is None:
            return
        self._logger.warning(msg)

    def exception(self, msg: str) -> None:
        if self._logger is None:
            return
        self._logger.exception(msg)

    def close(self):
        if self._logger is None:
            return
        self._logger.removeHandler(self.queue_handler)
        self.qlistner.stop()
        self._logger = None

def load_component(config: ConfigHelper) -> SimplyPrint:
    return SimplyPrint(config)



# Integration with Spoolman
#
# Copyright (C) 2023 Daniel Hultgren <daniel.cf.hultgren@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import asyncio
import logging
import re
import contextlib
import tornado.websocket as tornado_ws
from ..common import RequestType, HistoryFieldData
from ..utils import json_wrapper as jsonw
from typing import (
    TYPE_CHECKING,
    List,
    Dict,
    Any,
    Optional,
    Union,
    cast
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .http_client import HttpClient, HttpResponse
    from .database import MoonrakerDatabase
    from .announcements import Announcements
    from .klippy_apis import KlippyAPI as APIComp
    from .history import History
    from tornado.websocket import WebSocketClientConnection

DB_NAMESPACE = "moonraker"
ACTIVE_SPOOL_KEY = "spoolman.spool_id"

class SpoolManager:
    def __init__(self, config: ConfigHelper):
        self.server = config.get_server()
        self.eventloop = self.server.get_event_loop()
        self._get_spoolman_urls(config)
        self.sync_rate_seconds = config.getint("sync_rate", default=5, minval=1)
        self.report_timer = self.eventloop.register_timer(self.report_extrusion)
        self.pending_reports: Dict[int, float] = {}
        self.spoolman_ws: Optional[WebSocketClientConnection] = None
        self.connection_task: Optional[asyncio.Task] = None
        self.spool_check_task: Optional[asyncio.Task] = None
        self.ws_connected: bool = False
        self.reconnect_delay: float = 2.
        self.is_closing: bool = False
        self.spool_id: Optional[int] = None
        self._error_logged: bool = False
        self._highest_epos: float = 0
        self._current_extruder: str = "extruder"
        self.spool_history = HistoryFieldData(
            "spool_ids", "spoolman", "Spool IDs used", "collect",
            reset_callback=self._on_history_reset
        )
        history: History = self.server.lookup_component("history")
        history.register_auxiliary_field(self.spool_history)
        self.klippy_apis: APIComp = self.server.lookup_component("klippy_apis")
        self.http_client: HttpClient = self.server.lookup_component("http_client")
        self.database: MoonrakerDatabase = self.server.lookup_component("database")
        announcements: Announcements = self.server.lookup_component("announcements")
        announcements.register_feed("spoolman")
        self._register_notifications()
        self._register_listeners()
        self._register_endpoints()
        self.server.register_remote_method(
            "spoolman_set_active_spool", self.set_active_spool
        )

    def _get_spoolman_urls(self, config: ConfigHelper) -> None:
        orig_url = config.get('server')
        url_match = re.match(r"(?i:(?P<scheme>https?)://)?(?P<host>.+)", orig_url)
        if url_match is None:
            raise config.error(
                f"Section [spoolman], Option server: {orig_url}: Invalid URL format"
            )
        scheme = url_match["scheme"] or "http"
        host = url_match["host"].rstrip("/")
        ws_scheme = "wss" if scheme == "https" else "ws"
        self.spoolman_url = f"{scheme}://{host}/api"
        self.ws_url = f"{ws_scheme}://{host}/api/v1/spool"

    def _register_notifications(self):
        self.server.register_notification("spoolman:active_spool_set")
        self.server.register_notification("spoolman:spoolman_status_changed")

    def _register_listeners(self):
        self.server.register_event_handler(
            "server:klippy_ready", self._handle_klippy_ready
        )

    def _register_endpoints(self):
        self.server.register_endpoint(
            "/server/spoolman/spool_id",
            RequestType.GET | RequestType.POST,
            self._handle_spool_id_request,
        )
        self.server.register_endpoint(
            "/server/spoolman/proxy",
            RequestType.POST,
            self._proxy_spoolman_request,
        )
        self.server.register_endpoint(
            "/server/spoolman/status",
            RequestType.GET,
            self._handle_status_request,
        )

    def _on_history_reset(self) -> List[int]:
        if self.spool_id is None:
            return []
        return [self.spool_id]

    async def component_init(self) -> None:
        self.spool_id = await self.database.get_item(
            DB_NAMESPACE, ACTIVE_SPOOL_KEY, None
        )
        self.connection_task = self.eventloop.create_task(self._connect_websocket())

    async def _connect_websocket(self) -> None:
        log_connect: bool = True
        err_list: List[Exception] = []
        while not self.is_closing:
            if log_connect:
                logging.info(f"Connecting To Spoolman: {self.ws_url}")
                log_connect = False
            try:
                self.spoolman_ws = await tornado_ws.websocket_connect(
                    self.ws_url,
                    connect_timeout=5.,
                    ping_interval=20.,
                    ping_timeout=60.
                )
                setattr(self.spoolman_ws, "on_ping", self._on_ws_ping)
                cur_time = self.eventloop.get_loop_time()
                self._last_ping_received = cur_time
            except asyncio.CancelledError:
                raise
            except Exception as e:
                if len(err_list) < 10:
                    # Allow up to 10 unique errors.
                    for err in err_list:
                        if type(err) is type(e) and err.args == e.args:
                            break
                    else:
                        err_list.append(e)
                        verbose = self.server.is_verbose_enabled()
                        if verbose:
                            logging.exception("Failed to connect to Spoolman")
                        self.server.add_log_rollover_item(
                            "spoolman_connect", f"Failed to Connect to spoolman: {e}",
                            not verbose
                        )
            else:
                err_list = []
                self.ws_connected = True
                self._error_logged = False
                self.report_timer.start()
                self.server.add_log_rollover_item(
                    "spoolman_connect", "Connected to Spoolman Spool Manager"
                )
                if self.spool_id is not None:
                    self._cancel_spool_check_task()
                    coro = self._check_spool_deleted()
                    self.spool_check_task = self.eventloop.create_task(coro)
                self._send_status_notification()
                await self._read_messages()
                log_connect = True
            if not self.is_closing:
                await asyncio.sleep(self.reconnect_delay)

    async def _read_messages(self) -> None:
        message: Union[str, bytes, None]
        while self.spoolman_ws is not None:
            message = await self.spoolman_ws.read_message()
            if isinstance(message, str):
                self._decode_message(message)
            elif message is None:
                self.report_timer.stop()
                self.ws_connected = False
                cur_time = self.eventloop.get_loop_time()
                ping_time: float = cur_time - self._last_ping_received
                reason = code = None
                if self.spoolman_ws is not None:
                    reason = self.spoolman_ws.close_reason
                    code = self.spoolman_ws.close_code
                logging.info(
                    f"Spoolman Disconnected - Code: {code}, Reason: {reason}, "
                    f"Server Ping Time Elapsed: {ping_time}"
                )
                self.spoolman_ws = None
                if not self.is_closing:
                    self._send_status_notification()
                break

    def _decode_message(self, message: str) -> None:
        event: Dict[str, Any] = jsonw.loads(message)
        if event.get("resource") != "spool":
            return
        if self.spool_id is not None and event.get("type") == "deleted":
            payload: Dict[str, Any] = event.get("payload", {})
            if payload.get("id") == self.spool_id:
                self.pending_reports.pop(self.spool_id, None)
                self.set_active_spool(None)

    def _cancel_spool_check_task(self) -> None:
        if self.spool_check_task is None or self.spool_check_task.done():
            return
        self.spool_check_task.cancel()

    async def _check_spool_deleted(self) -> None:
        if self.spool_id is not None:
            response = await self.http_client.get(
                f"{self.spoolman_url}/v1/spool/{self.spool_id}",
                connect_timeout=1., request_timeout=2.
            )
            if response.status_code == 404:
                logging.info(f"Spool ID {self.spool_id} not found, setting to None")
                self.pending_reports.pop(self.spool_id, None)
                self.set_active_spool(None)
            elif response.has_error():
                err_msg = self._get_response_error(response)
                logging.info(f"Attempt to check spool status failed: {err_msg}")
            else:
                logging.info(f"Found Spool ID {self.spool_id} on spoolman instance")
        self.spool_check_task = None

    def connected(self) -> bool:
        return self.ws_connected

    def _on_ws_ping(self, data: bytes = b"") -> None:
        self._last_ping_received = self.eventloop.get_loop_time()

    async def _handle_klippy_ready(self) -> None:
        result: Dict[str, Dict[str, Any]]
        result = await self.klippy_apis.subscribe_objects(
            {"toolhead": ["position", "extruder"]}, self._handle_status_update, {}
        )
        toolhead = result.get("toolhead", {})
        self._current_extruder = toolhead.get("extruder", "extruder")
        initial_e_pos = toolhead.get("position", [None]*4)[3]
        logging.debug(f"Initial epos: {initial_e_pos}")
        if initial_e_pos is not None:
            self._highest_epos = initial_e_pos
        else:
            logging.error("Spoolman integration unable to subscribe to epos")
            raise self.server.error("Unable to subscribe to e position")

    def _get_response_error(self, response: HttpResponse) -> str:
        err_msg = f"HTTP error: {response.status_code} {response.error}"
        with contextlib.suppress(Exception):
            msg: Optional[str] = cast(dict, response.json())["message"]
            err_msg += f", Spoolman message: {msg}"
        return err_msg

    def _handle_status_update(self, status: Dict[str, Any], _: float) -> None:
        toolhead: Optional[Dict[str, Any]] = status.get("toolhead")
        if toolhead is None:
            return
        epos: float = toolhead.get("position", [0, 0, 0, self._highest_epos])[3]
        extr = toolhead.get("extruder", self._current_extruder)
        if extr != self._current_extruder:
            self._highest_epos = epos
            self._current_extruder = extr
        elif epos > self._highest_epos:
            if self.spool_id is not None:
                self._add_extrusion(self.spool_id, epos - self._highest_epos)
            self._highest_epos = epos

    def _add_extrusion(self, spool_id: int, used_length: float) -> None:
        if spool_id in self.pending_reports:
            self.pending_reports[spool_id] += used_length
        else:
            self.pending_reports[spool_id] = used_length

    def set_active_spool(self, spool_id: Union[int, None]) -> None:
        assert spool_id is None or isinstance(spool_id, int)
        if self.spool_id == spool_id:
            logging.info(f"Spool ID already set to: {spool_id}")
            return
        self.spool_history.tracker.update(spool_id)
        self.spool_id = spool_id
        self.database.insert_item(DB_NAMESPACE, ACTIVE_SPOOL_KEY, spool_id)
        self.server.send_event(
            "spoolman:active_spool_set", {"spool_id": spool_id}
        )
        logging.info(f"Setting active spool to: {spool_id}")

    async def report_extrusion(self, eventtime: float) -> float:
        if not self.ws_connected:
            return eventtime + self.sync_rate_seconds
        pending_reports = self.pending_reports
        self.pending_reports = {}
        for spool_id, used_length in pending_reports.items():
            if not self.ws_connected:
                self._add_extrusion(spool_id, used_length)
                continue
            logging.debug(
                f"Sending spool usage: ID: {spool_id}, Length: {used_length:.3f}mm"
            )
            response = await self.http_client.request(
                method="PUT",
                url=f"{self.spoolman_url}/v1/spool/{spool_id}/use",
                body={"use_length": used_length}
            )
            if response.has_error():
                if response.status_code == 404:
                    # Since the spool is deleted we can remove any pending reports
                    # added while waiting for the request
                    self.pending_reports.pop(spool_id, None)
                    if spool_id == self.spool_id:
                        logging.info(f"Spool ID {spool_id} not found, setting to None")
                        self.set_active_spool(None)
                else:
                    if not self._error_logged:
                        error_msg = self._get_response_error(response)
                        self._error_logged = True
                        logging.info(
                            f"Failed to update extrusion for spool id {spool_id}, "
                            f"received {error_msg}"
                        )
                    # Add missed reports back to pending reports for the next cycle
                    self._add_extrusion(spool_id, used_length)
                    continue
            self._error_logged = False
        return self.eventloop.get_loop_time() + self.sync_rate_seconds

    async def _handle_spool_id_request(self, web_request: WebRequest):
        if web_request.get_request_type() == RequestType.POST:
            spool_id = web_request.get_int("spool_id", None)
            self.set_active_spool(spool_id)
        # For GET requests we will simply return the spool_id
        return {"spool_id": self.spool_id}

    async def _proxy_spoolman_request(self, web_request: WebRequest):
        method = web_request.get_str("request_method")
        path = web_request.get_str("path")
        query = web_request.get_str("query", None)
        body = web_request.get("body", None)
        use_v2_response = web_request.get_boolean("use_v2_response", False)
        if method not in {"GET", "POST", "PUT", "PATCH", "DELETE"}:
            raise self.server.error(f"Invalid HTTP method: {method}")
        if body is not None and method == "GET":
            raise self.server.error("GET requests cannot have a body")
        if len(path) < 4 or path[:4] != "/v1/":
            raise self.server.error(
                "Invalid path, must start with the API version, e.g. /v1"
            )
        query = f"?{query}" if query is not None else ""
        full_url = f"{self.spoolman_url}{path}{query}"
        if not self.ws_connected:
            if not use_v2_response:
                raise self.server.error("Spoolman server not available", 503)
            return {
                "response": None,
                "error": {
                    "status_code": 503,
                    "message": "Spoolman server not available"
                }
            }
        logging.debug(f"Proxying {method} request to {full_url}")
        response = await self.http_client.request(
            method=method,
            url=full_url,
            body=body,
        )
        if not use_v2_response:
            response.raise_for_status()
            return response.json()
        if response.has_error():
            msg: str = str(response.error or "")
            with contextlib.suppress(Exception):
                spoolman_msg = cast(dict, response.json()).get("message", msg)
                msg = spoolman_msg
            return {
                "response": None,
                "error": {
                    "status_code": response.status_code,
                    "message": msg
                }
            }
        else:
            return {
                "response": response.json(),
                "response_headers": dict(response.headers.items()),
                "error": None
            }

    async def _handle_status_request(self, web_request: WebRequest) -> Dict[str, Any]:
        pending: List[Dict[str, Any]] = [
            {"spool_id": sid, "filament_used": used} for sid, used in
            self.pending_reports.items()
        ]
        return {
            "spoolman_connected": self.ws_connected,
            "pending_reports": pending,
            "spool_id": self.spool_id
        }

    def _send_status_notification(self) -> None:
        self.server.send_event(
            "spoolman:spoolman_status_changed",
            {"spoolman_connected": self.ws_connected}
        )

    async def close(self):
        self.is_closing = True
        self.report_timer.stop()
        if self.spoolman_ws is not None:
            self.spoolman_ws.close(1001, "Moonraker Shutdown")
        self._cancel_spool_check_task()
        if self.connection_task is None or self.connection_task.done():
            return
        try:
            await asyncio.wait_for(self.connection_task, 2.)
        except asyncio.TimeoutError:
            pass

def load_component(config: ConfigHelper) -> SpoolManager:
    return SpoolManager(config)



# Package entry point for Moonraker
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from .server import main

main()



# Wrapper for msgspec with stdlib fallback
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import os
import contextlib
from typing import Any, Union, TYPE_CHECKING

if TYPE_CHECKING:
    def dumps(obj: Any) -> bytes: ...  # type: ignore # noqa: E704
    def loads(data: Union[str, bytes, bytearray]) -> Any: ...  # noqa: E704

MSGSPEC_ENABLED = False
_msgspc_var = os.getenv("MOONRAKER_ENABLE_MSGSPEC", "y").lower()
if _msgspc_var in ["y", "yes", "true"]:
    with contextlib.suppress(ImportError):
        import msgspec
        from msgspec import DecodeError as JSONDecodeError
        encoder = msgspec.json.Encoder()
        decoder = msgspec.json.Decoder()
        dumps = encoder.encode  # noqa: F811
        loads = decoder.decode  # noqa: F811
        MSGSPEC_ENABLED = True
if not MSGSPEC_ENABLED:
    import json
    from json import JSONDecodeError  # type: ignore # noqa: F401,F811
    loads = json.loads  # type: ignore

    def dumps(obj) -> bytes:  # type: ignore # noqa: F811
        return json.dumps(obj).encode("utf-8")



# Centralized webcam configuration
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import re
import ipaddress
import socket
import uuid
import logging
from ..common import RequestType
from typing import (
    TYPE_CHECKING,
    Optional,
    Dict,
    List,
    Any,
)

if TYPE_CHECKING:
    from asyncio import Future
    from ..server import Server
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .database import MoonrakerDatabase
    from .machine import Machine
    from .shell_command import ShellCommandFactory
    from .http_client import HttpClient

# This provides a mapping of fields defined by Moonraker to fields
# defined by the database.
CAM_FIELDS = {
    "name": "name", "service": "service", "target_fps": "targetFps",
    "stream_url": "urlStream", "snapshot_url": "urlSnapshot",
    "flip_horizontal": "flipX", "flip_vertical": "flipY",
    "enabled": "enabled", "target_fps_idle": "targetFpsIdle",
    "aspect_ratio": "aspectRatio", "icon": "icon"
}

class WebcamManager:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.webcams: Dict[str, WebCam] = {}
        # parse user configured webcams
        prefix_sections = config.get_prefix_sections("webcam ")
        for section in prefix_sections:
            cam_cfg = config[section]
            webcam = WebCam.from_config(cam_cfg)
            self.webcams[webcam.name] = webcam

        self.server.register_endpoint(
            "/server/webcams/list", RequestType.GET, self._handle_webcam_list
        )
        self.server.register_endpoint(
            "/server/webcams/item", RequestType.all(),
            self._handle_webcam_request
        )
        self.server.register_endpoint(
            "/server/webcams/test", RequestType.POST, self._handle_webcam_test
        )
        self.server.register_notification("webcam:webcams_changed")
        self.server.register_event_handler(
            "machine:public_ip_changed", self._set_default_host_ip
        )

    async def component_init(self) -> None:
        machine: Machine = self.server.lookup_component("machine")
        if machine.public_ip:
            self._set_default_host_ip(machine.public_ip)
        all_uids = [wc.uid for wc in self.webcams.values()]
        db: MoonrakerDatabase = self.server.lookup_component("database")
        db_cams: Dict[str, Dict[str, Any]] = await db.get_item("webcams", default={})
        ro_info: List[str] = []
        # Process configured cams
        for uid, cam_data in db_cams.items():
            try:
                cam_data["uid"] = uid
                webcam = WebCam.from_database(self.server, cam_data)
                if uid in all_uids:
                    # Unlikely but possible collision between random UUID4
                    # and UUID5 generated from a configured webcam.
                    await db.delete_item("webcams", uid)
                    webcam.uid = self._get_guaranteed_uuid()
                    await self._save_cam(webcam, False)
                    ro_info.append(f"Detected webcam UID collision: {uid}")
                all_uids.append(webcam.uid)
                if webcam.name in self.webcams:
                    ro_info.append(
                        f"Detected webcam name collision: {webcam.name}, uuid: "
                        f"{uid}.  This camera will be ignored."
                    )
                    continue
                self.webcams[webcam.name] = webcam
            except Exception:
                logging.exception("Failed to process webcam from db")
                continue
        if ro_info:
            self.server.add_log_rollover_item("webcam", "\n".join(ro_info))

    def _set_default_host_ip(self, ip: str) -> None:
        default_host = "http://127.0.0.1"
        if ip:
            try:
                addr = ipaddress.ip_address(ip)
            except Exception:
                logging.debug(f"Invalid IP Recd: {ip}")
            else:
                if addr.version == 6:
                    default_host = f"http://[{addr}]"
                else:
                    default_host = f"http://{addr}"
        WebCam.set_default_host(default_host)
        logging.info(f"Default public webcam address set: {default_host}")

    def get_webcams(self) -> Dict[str, WebCam]:
        return self.webcams

    def _list_webcams(self) -> List[Dict[str, Any]]:
        return [wc.as_dict() for wc in self.webcams.values()]

    def _save_cam(self, webcam: WebCam, save_local: bool = True) -> Future:
        if save_local:
            self.webcams[webcam.name] = webcam
        cam_data: Dict[str, Any] = {}
        for mfield, dbfield in CAM_FIELDS.items():
            cam_data[dbfield] = getattr(webcam, mfield)
        cam_data["location"] = webcam.location
        cam_data["rotation"] = webcam.rotation
        cam_data["extra_data"] = webcam.extra_data
        db: MoonrakerDatabase = self.server.lookup_component("database")
        return db.insert_item("webcams", webcam.uid, cam_data)

    def _delete_cam(self, webcam: WebCam) -> Future:
        db: MoonrakerDatabase = self.server.lookup_component("database")
        self.webcams.pop(webcam.name, None)
        return db.delete_item("webcams", webcam.uid)

    def _get_guaranteed_uuid(self) -> str:
        cur_uids = [wc.uid for wc in self.webcams.values()]
        while True:
            uid = str(uuid.uuid4())
            if uid not in cur_uids:
                break
        return uid

    def get_cam_by_uid(self, uid: str) -> WebCam:
        for cam in self.webcams.values():
            if cam.uid == uid:
                return cam
        raise self.server.error(f"Webcam with UID {uid} not found", 404)

    def _lookup_camera(
        self, web_request: WebRequest, required: bool = True
    ) -> Optional[WebCam]:
        args = web_request.get_args()
        if "uid" in args:
            return self.get_cam_by_uid(web_request.get_str("uid"))
        name = web_request.get_str("name")
        webcam = self.webcams.get(name, None)
        if required and webcam is None:
            raise self.server.error(f"Webcam {name} not found", 404)
        return webcam

    async def _handle_webcam_request(self, web_request: WebRequest) -> Dict[str, Any]:
        req_type = web_request.get_request_type()
        webcam = self._lookup_camera(web_request, req_type != RequestType.POST)
        webcam_data: Dict[str, Any] = {}
        if req_type == RequestType.GET:
            assert webcam is not None
            webcam_data = webcam.as_dict()
        elif req_type == RequestType.POST:
            if webcam is not None:
                if webcam.source == "config":
                    raise self.server.error(
                        f"Cannot overwrite webcam '{webcam.name}' sourced from "
                        "Moonraker configuration"
                    )
                new_name = web_request.get_str("name", None)
                if new_name is not None and webcam.name != new_name:
                    if new_name in self.webcams:
                        raise self.server.error(
                            f"Cannot rename webcam from '{webcam.name}' to "
                            f"'{new_name}'.  Webcam with requested name '{new_name}' "
                            "already exists."
                        )
                    self.webcams.pop(webcam.name, None)
                webcam.update(web_request)
            else:
                uid = self._get_guaranteed_uuid()
                webcam = WebCam.from_web_request(self.server, web_request, uid)
            await self._save_cam(webcam)
            webcam_data = webcam.as_dict()
        elif req_type == RequestType.DELETE:
            assert webcam is not None
            if webcam.source == "config":
                raise self.server.error(
                    f"Cannot delete webcam '{webcam.name}' sourced from "
                    "Moonraker configuration"
                )
            webcam_data = webcam.as_dict()
            self._delete_cam(webcam)
        if req_type != RequestType.GET:
            self.server.send_event(
                "webcam:webcams_changed", {"webcams": self._list_webcams()}
            )
        return {"webcam": webcam_data}

    async def _handle_webcam_list(self, web_request: WebRequest) -> Dict[str, Any]:
        return {"webcams": self._list_webcams()}

    async def _handle_webcam_test(self, web_request: WebRequest) -> Dict[str, Any]:
        client: HttpClient = self.server.lookup_component("http_client")
        webcam = self._lookup_camera(web_request)
        assert webcam is not None
        result: Dict[str, Any] = {
            "name": webcam.name,
            "snapshot_reachable": False
        }
        for img_type in ["snapshot", "stream"]:
            try:
                func = getattr(webcam, f"get_{img_type}_url")
                result[f"{img_type}_url"] = await func(True)
            except Exception:
                logging.exception(f"Error Processing {img_type} url")
                result[f"{img_type}_url"] = ""
        url: str = result["snapshot_url"]
        if url.startswith("http"):
            ret = await client.get(url, connect_timeout=1., request_timeout=1.)
            result["snapshot_reachable"] = not ret.has_error()
        return result


class WebCam:
    _default_host: str = "http://127.0.0.1"
    _protected_fields: List[str] = ["source", "uid"]
    def __init__(self, server: Server, **kwargs) -> None:
        self._server = server
        self.name: str = kwargs["name"]
        self.enabled: bool = kwargs["enabled"]
        self.icon: str = kwargs["icon"]
        self.aspect_ratio: str = kwargs["aspect_ratio"]
        self.target_fps: int = kwargs["target_fps"]
        self.target_fps_idle: int = kwargs["target_fps_idle"]
        self.location: str = kwargs["location"]
        self.service: str = kwargs["service"]
        self.stream_url: str = kwargs["stream_url"]
        self.snapshot_url: str = kwargs["snapshot_url"]
        self.flip_horizontal: bool = kwargs["flip_horizontal"]
        self.flip_vertical: bool = kwargs["flip_vertical"]
        self.rotation: int = kwargs["rotation"]
        self.source: str = kwargs["source"]
        self.extra_data: Dict[str, Any] = kwargs.get("extra_data", {})
        self.uid: str = kwargs["uid"]
        if self.rotation not in [0, 90, 180, 270]:
            raise server.error(f"Invalid value for 'rotation': {self.rotation}")
        prefix, sep, postfix = self.aspect_ratio.partition(":")
        if not (prefix.isdigit() and sep == ":" and postfix.isdigit()):
            raise server.error(
                f"Invalid value for 'aspect_ratio': {self.aspect_ratio}"
            )

    def as_dict(self):
        return {k: v for k, v in self.__dict__.items() if k[0] != "_"}

    async def get_stream_url(self, convert_local: bool = False) -> str:
        return await self._get_url(self.stream_url, convert_local)

    async def get_snapshot_url(self, convert_local: bool = False) -> str:
        return await self._get_url(self.snapshot_url, convert_local)

    async def _get_url(self, url: str, convert_local: bool) -> str:
        if not url:
            raise self._server.error("Empty URL Provided")
        match = re.match(r"\w+://[^/]+", url)
        if match is None:
            # assume a partial URL on the default host
            url = f"{self._default_host}/{url.lstrip('/')}"
        if not convert_local:
            return url
        return await self.convert_local(url)

    def _get_local_ips(self) -> List[str]:
        all_ips: List[str] = []
        machine: Machine = self._server.lookup_component("machine")
        sys_info = machine.get_system_info()
        network = sys_info.get("network", {})
        iface: Dict[str, Any]
        for iface in network.values():
            addresses: List[Dict[str, Any]] = iface["ip_addresses"]
            for addr_info in addresses:
                all_ips.append(addr_info["address"])
        return all_ips

    async def convert_local(self, url: str) -> str:
        match = re.match(r"(\w+)://([^/]+)(/.*)?", url)
        if match is None:
            return url
        scheme = match.group(1)
        addr = match.group(2)
        fragment = match.group(3)
        if fragment is None:
            fragment = ""
        if addr[0] == "[":
            # ipv6 address
            addr_match = re.match(r"\[(.+)\](:\d+)?", addr)
        else:
            # ipv4 address or hostname
            addr_match = re.match(r"([^:]+)(:\d+)?", addr)
        if addr_match is None:
            return url
        addr = addr_match.group(1)
        port: Optional[str] = addr_match.group(2)
        default_ports = {"http": "80", "https": "443", "rtsp": "554"}
        if port is None:
            if scheme not in default_ports:
                return url
            port = default_ports[scheme]
        else:
            port = port.lstrip(":")
        # attempt to convert hostname to IP
        try:
            eventloop = self._server.get_event_loop()
            addr_info = await eventloop.run_in_thread(
                socket.getaddrinfo, addr, int(port)
            )
            if addr_info:
                addr = addr_info[0][4][0]
        except Exception:
            pass
        try:
            ip = ipaddress.ip_address(addr)
        except Exception:
            # Invalid IP, can't convert.
            return url
        else:
            if ip.is_loopback:
                return url
            # Check to see if this ip address is on the local machine
            if addr not in self._get_local_ips():
                return url
        scmd: ShellCommandFactory
        scmd = self._server.lookup_component("shell_command")
        try:
            # Use the ss command to list all tcp ports
            resp: str = await scmd.exec_cmd("ss -ltn")
            lines = resp.split("\n")[1:]
            for line in lines:
                parts = line.strip().split()
                if len(parts) < 5:
                    continue
                laddr, lport = parts[3].split(":")
                if lport == port:
                    if laddr == "[::]":
                        return f"{scheme}://[::1]:{port}{fragment}"
                    elif laddr == "0.0.0.0":
                        return f"{scheme}://127.0.0.1:{port}{fragment}"
        except scmd.error:
            pass
        return url

    def update(self, web_request: WebRequest) -> None:
        valid_fields = [
            f for f in self.__dict__.keys() if f[0] != "_"
            and f not in self._protected_fields
        ]
        for field in web_request.get_args().keys():
            if field not in valid_fields:
                continue
            try:
                attr_type = type(getattr(self, field))
            except AttributeError:
                continue
            if attr_type is bool:
                val: Any = web_request.get_boolean(field)
            elif attr_type is int:
                val = web_request.get_int(field)
            elif attr_type is float:
                val = web_request.get_float(field)
            elif attr_type is str:
                val = web_request.get_str(field)
            else:
                val = web_request.get(field)
            setattr(self, field, val)

    @staticmethod
    def set_default_host(host: str) -> None:
        WebCam._default_host = host

    @classmethod
    def from_config(cls, config: ConfigHelper) -> WebCam:
        server = config.get_server()
        name = config.get_name().split(maxsplit=1)[-1]
        ns = uuid.UUID(server.get_app_args()["instance_uuid"])
        try:
            return cls(
                server,
                name=name,
                enabled=config.getboolean("enabled", True),
                icon=config.get("icon", "mdiWebcam"),
                aspect_ratio=config.get("aspect_ratio", "4:3"),
                target_fps=config.getint("target_fps", 15),
                target_fps_idle=config.getint("target_fps_idle", 5),
                location=config.get("location", "printer"),
                service=config.get("service", "mjpegstreamer"),
                stream_url=config.get("stream_url"),
                snapshot_url=config.get("snapshot_url", ""),
                flip_horizontal=config.getboolean("flip_horizontal", False),
                flip_vertical=config.getboolean("flip_vertical", False),
                rotation=config.getint("rotation", 0),
                source="config",
                uid=str(uuid.uuid5(ns, f"moonraker.webcam.{name}"))
            )
        except server.error as err:
            raise config.error(str(err)) from err

    @classmethod
    def from_web_request(
        cls, server: Server, web_request: WebRequest, uid: str
    ) -> WebCam:
        name = web_request.get_str("name")
        return cls(
            server,
            name=name,
            enabled=web_request.get_boolean("enabled", True),
            icon=web_request.get_str("icon", "mdiWebcam"),
            aspect_ratio=web_request.get_str("aspect_ratio", "4:3"),
            target_fps=web_request.get_int("target_fps", 15),
            target_fps_idle=web_request.get_int("target_fps_idle", 5),
            location=web_request.get_str("location", "printer"),
            service=web_request.get_str("service", "mjpegstreamer"),
            stream_url=web_request.get_str("stream_url"),
            snapshot_url=web_request.get_str("snapshot_url", ""),
            flip_horizontal=web_request.get_boolean("flip_horizontal", False),
            flip_vertical=web_request.get_boolean("flip_vertical", False),
            rotation=web_request.get_int("rotation", 0),
            source="database",
            extra_data=web_request.get("extra_data", {}),
            uid=uid
        )

    @classmethod
    def from_database(cls, server: Server, cam_data: Dict[str, Any]) -> WebCam:
        return cls(
            server,
            name=str(cam_data["name"]),
            enabled=bool(cam_data.get("enabled", True)),
            icon=str(cam_data.get("icon", "mdiWebcam")),
            aspect_ratio=str(cam_data.get("aspectRatio", "4:3")),
            target_fps=int(cam_data.get("targetFps", 15)),
            target_fps_idle=int(cam_data.get("targetFpsIdle", 5)),
            location=str(cam_data.get("location", "printer")),
            service=str(cam_data.get("service", "mjpegstreamer")),
            stream_url=str(cam_data.get("urlStream", "")),
            snapshot_url=str(cam_data.get("urlSnapshot", "")),
            flip_horizontal=bool(cam_data.get("flipX", False)),
            flip_vertical=bool(cam_data.get("flipY", False)),
            rotation=int(cam_data.get("rotation", cam_data.get("rotate", 0))),
            source="database",
            extra_data=cam_data.get("extra_data", {}),
            uid=cam_data["uid"]
        )

def load_component(config: ConfigHelper) -> WebcamManager:
    return WebcamManager(config)



# API Key Based Authorization
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import asyncio
import base64
import uuid
import hashlib
import secrets
import os
import time
import datetime
import ipaddress
import re
import socket
import logging
from tornado.web import HTTPError
from libnacl.sign import Signer, Verifier
from ..utils import json_wrapper as jsonw
from ..common import RequestType, TransportType, SqlTableDefinition, UserInfo

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
    Optional,
    Union,
    Dict,
    List,
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .websockets import WebsocketManager
    from tornado.httputil import HTTPServerRequest
    from .database import MoonrakerDatabase as DBComp
    from .database import DBProviderWrapper
    from .ldap import MoonrakerLDAP
    IPAddr = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
    IPNetwork = Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
    OneshotToken = Tuple[IPAddr, Optional[UserInfo], asyncio.Handle]

# Helpers for base64url encoding and decoding
def base64url_encode(data: bytes) -> bytes:
    return base64.urlsafe_b64encode(data).rstrip(b"=")

def base64url_decode(data: str) -> bytes:
    pad_cnt = len(data) % 4
    if pad_cnt:
        data += "=" * (4 - pad_cnt)
    return base64.urlsafe_b64decode(data)


ONESHOT_TIMEOUT = 5
TRUSTED_CONNECTION_TIMEOUT = 3600
FQDN_CACHE_TIMEOUT = 84000
PRUNE_CHECK_TIME = 300.

USER_TABLE = "authorized_users"
AUTH_SOURCES = ["moonraker", "ldap"]
HASH_ITER = 100000
API_USER = "_API_KEY_USER_"
TRUSTED_USER = "_TRUSTED_USER_"
RESERVED_USERS = [API_USER, TRUSTED_USER]
JWT_EXP_TIME = datetime.timedelta(hours=1)
JWT_HEADER = {
    'alg': "EdDSA",
    'typ': "JWT"
}

class UserSqlDefinition(SqlTableDefinition):
    name = USER_TABLE
    prototype = (
        f"""
        {USER_TABLE} (
            username TEXT PRIMARY KEY NOT NULL,
            password TEXT NOT NULL,
            created_on REAL NOT NULL,
            salt TEXT NOT NULL,
            source TEXT NOT NULL,
            jwt_secret TEXT,
            jwk_id TEXT,
            groups pyjson
        )
        """
    )
    version = 1

    def migrate(self, last_version: int, db_provider: DBProviderWrapper) -> None:
        if last_version == 0:
            users: Dict[str, Dict[str, Any]]
            users = db_provider.get_namespace("authorized_users")
            api_user = users.pop(API_USER, {})
            if not isinstance(api_user, dict):
                api_user = {}
            user_vals: List[Tuple[Any, ...]] = [
                UserInfo(
                    username=API_USER,
                    password=api_user.get("api_key", uuid.uuid4().hex),
                    created_on=api_user.get("created_on", time.time())
                ).as_tuple()
            ]
            for key, user in users.items():
                if not isinstance(user, dict):
                    logging.info(
                        f"Auth migration, skipping invalid value: {key} {user}"
                    )
                    continue
                user_vals.append(UserInfo(**user).as_tuple())
            placeholders = ",".join("?" * len(user_vals[0]))
            conn = db_provider.connection
            with conn:
                conn.executemany(
                    f"INSERT OR IGNORE INTO {USER_TABLE} VALUES({placeholders})",
                    user_vals
                )
            db_provider.wipe_local_namespace("authorized_users")

class Authorization:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.login_timeout = config.getint('login_timeout', 90)
        self.force_logins = config.getboolean('force_logins', False)
        self.default_source = config.get('default_source', "moonraker").lower()
        self.enable_api_key = config.getboolean('enable_api_key', True)
        self.max_logins = config.getint("max_login_attempts", None, above=0)
        self.failed_logins: Dict[IPAddr, int] = {}
        self.fqdn_cache: Dict[IPAddr, Dict[str, Any]] = {}
        if self.default_source not in AUTH_SOURCES:
            self.server.add_warning(
                "[authorization]: option 'default_source' - Invalid "
                f"value '{self.default_source}', falling back to "
                "'moonraker'."
            )
            self.default_source = "moonraker"
        self.ldap: Optional[MoonrakerLDAP] = None
        if config.has_section("ldap"):
            self.ldap = self.server.load_component(config, "ldap", None)
        if self.default_source == "ldap" and self.ldap is None:
            self.server.add_warning(
                "[authorization]: Option 'default_source' set to 'ldap',"
                " however [ldap] section failed to load or not configured"
            )
        database: DBComp = self.server.lookup_component('database')
        self.user_table = database.register_table(UserSqlDefinition())
        self.users: Dict[str, UserInfo] = {}
        self.api_key = uuid.uuid4().hex
        hi = self.server.get_host_info()
        self.issuer = f"http://{hi['hostname']}:{hi['port']}"
        self.public_jwks: Dict[str, Dict[str, Any]] = {}
        self.trusted_users: Dict[IPAddr, Dict[str, Any]] = {}
        self.oneshot_tokens: Dict[str, OneshotToken] = {}

        # Get allowed cors domains
        self.cors_domains: List[str] = []
        for domain in config.getlist('cors_domains', []):
            bad_match = re.search(r"^.+\.[^:]*\*", domain)
            if bad_match is not None:
                self.server.add_warning(
                    f"[authorization]: Unsafe domain '{domain}' in option "
                    f"'cors_domains'. Wildcards are not permitted in the"
                    " top level domain."
                )
                continue
            if domain.endswith("/"):
                self.server.add_warning(
                    f"[authorization]: Invalid domain '{domain}' in option "
                    "'cors_domains'.  Domain's cannot contain a trailing "
                    "slash."
                )
            else:
                self.cors_domains.append(
                    domain.replace(".", "\\.").replace("*", ".*"))

        # Get Trusted Clients
        self.trusted_ips: List[IPAddr] = []
        self.trusted_ranges: List[IPNetwork] = []
        self.trusted_domains: List[str] = []
        for val in config.getlist('trusted_clients', []):
            # Check IP address
            try:
                tc = ipaddress.ip_address(val)
            except ValueError:
                pass
            else:
                self.trusted_ips.append(tc)
                continue
            # Check ip network
            try:
                tn = ipaddress.ip_network(val)
            except ValueError as e:
                if "has host bits set" in str(e):
                    self.server.add_warning(
                        f"[authorization]: Invalid CIDR expression '{val}' "
                        "in option 'trusted_clients'")
                    continue
                pass
            else:
                self.trusted_ranges.append(tn)
                continue
            # Check hostname
            match = re.match(r"([a-z0-9]+(-[a-z0-9]+)*\.?)+[a-z]{2,}$", val)
            if match is not None:
                self.trusted_domains.append(val.lower())
            else:
                self.server.add_warning(
                    f"[authorization]: Invalid domain name '{val}' "
                    "in option 'trusted_clients'")

        t_clients = "\n".join(
            [str(ip) for ip in self.trusted_ips] +
            [str(rng) for rng in self.trusted_ranges] +
            self.trusted_domains)
        c_domains = "\n".join(self.cors_domains)

        logging.info(
            f"Authorization Configuration Loaded\n"
            f"Trusted Clients:\n{t_clients}\n"
            f"CORS Domains:\n{c_domains}")

        eventloop = self.server.get_event_loop()
        self.prune_timer = eventloop.register_timer(
            self._prune_conn_handler)

        # Register Authorization Endpoints
        self.server.register_endpoint(
            "/access/login", RequestType.POST, self._handle_login,
            transports=TransportType.HTTP | TransportType.WEBSOCKET,
            auth_required=False
        )
        self.server.register_endpoint(
            "/access/logout", RequestType.POST, self._handle_logout,
            transports=TransportType.HTTP | TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/access/refresh_jwt", RequestType.POST, self._handle_refresh_jwt,
            transports=TransportType.HTTP | TransportType.WEBSOCKET,
            auth_required=False
        )
        self.server.register_endpoint(
            "/access/user", RequestType.all(), self._handle_user_request,
            transports=TransportType.HTTP | TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/access/users/list", RequestType.GET, self._handle_list_request,
            transports=TransportType.HTTP | TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/access/user/password", RequestType.POST, self._handle_password_reset,
            transports=TransportType.HTTP | TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/access/api_key", RequestType.GET | RequestType.POST,
            self._handle_apikey_request,
            transports=TransportType.HTTP | TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/access/oneshot_token", RequestType.GET, self._handle_oneshot_request,
            transports=TransportType.HTTP | TransportType.WEBSOCKET
        )
        self.server.register_endpoint(
            "/access/info", RequestType.GET, self._handle_info_request,
            transports=TransportType.HTTP | TransportType.WEBSOCKET,
            auth_required=False
        )
        wsm: WebsocketManager = self.server.lookup_component("websockets")
        wsm.register_notification("authorization:user_created")
        wsm.register_notification(
            "authorization:user_deleted", event_type="logout"
        )
        wsm.register_notification(
            "authorization:user_logged_out", event_type="logout"
        )

    async def component_init(self) -> None:
        # Populate users from database
        cursor = await self.user_table.execute(f"SELECT * FROM {USER_TABLE}")
        self.users = {row[0]: UserInfo(**dict(row)) for row in await cursor.fetchall()}
        need_sync = self._initialize_users()
        if need_sync:
            await self._sync_user_table()
        self.prune_timer.start(delay=PRUNE_CHECK_TIME)

    async def _sync_user(self, username: str) -> None:
        user = self.users[username]
        vals = user.as_tuple()
        placeholders = ",".join("?" * len(vals))
        async with self.user_table as tx:
            await tx.execute(
                f"REPLACE INTO {USER_TABLE} VALUES({placeholders})", vals
            )

    async def _sync_user_table(self) -> None:
        async with self.user_table as tx:
            await tx.execute(f"DELETE FROM {USER_TABLE}")
            user_vals: List[Tuple[Any, ...]]
            user_vals = [user.as_tuple() for user in self.users.values()]
            if not user_vals:
                return
            placeholders = ",".join("?" * len(user_vals[0]))
            await tx.executemany(
                f"INSERT INTO {USER_TABLE} VALUES({placeholders})", user_vals
            )

    def _initialize_users(self) -> bool:
        need_sync = False
        api_user: Optional[UserInfo] = self.users.get(API_USER, None)
        if api_user is None:
            need_sync = True
            self.users[API_USER] = UserInfo(username=API_USER, password=self.api_key)
        else:
            self.api_key = api_user.password
        for username, user_info in list(self.users.items()):
            if username == API_USER:
                continue
            # generate jwks for valid users
            if user_info.jwt_secret is not None:
                try:
                    priv_key = self._load_private_key(user_info.jwt_secret)
                    jwk_id = user_info.jwk_id
                    assert jwk_id is not None
                except (self.server.error, KeyError, AssertionError):
                    logging.info("Invalid jwk found for user, removing")
                    user_info.jwt_secret = None
                    user_info.jwk_id = None
                    self.users[username] = user_info
                    need_sync = True
                    continue
                self.public_jwks[jwk_id] = self._generate_public_jwk(priv_key)
        return need_sync

    async def _handle_apikey_request(self, web_request: WebRequest) -> str:
        if web_request.get_request_type() == RequestType.POST:
            self.api_key = uuid.uuid4().hex
            self.users[API_USER].password = self.api_key
            await self._sync_user(API_USER)
        return self.api_key

    async def _handle_oneshot_request(self, web_request: WebRequest) -> str:
        ip = web_request.get_ip_address()
        assert ip is not None
        user_info = web_request.get_current_user()
        return self.get_oneshot_token(ip, user_info)

    async def _handle_login(self, web_request: WebRequest) -> Dict[str, Any]:
        ip = web_request.get_ip_address()
        if ip is not None and self.check_logins_maxed(ip):
            raise HTTPError(
                401, "Unauthorized, Maximum Login Attempts Reached"
            )
        try:
            ret = await self._login_jwt_user(web_request)
        except asyncio.CancelledError:
            raise
        except Exception:
            if ip is not None:
                failed = self.failed_logins.get(ip, 0)
                self.failed_logins[ip] = failed + 1
            raise
        if ip is not None:
            self.failed_logins.pop(ip, None)
        return ret

    async def _handle_logout(self, web_request: WebRequest) -> Dict[str, str]:
        user_info = web_request.get_current_user()
        if user_info is None:
            raise self.server.error("No user logged in")
        username: str = user_info.username
        if username in RESERVED_USERS:
            raise self.server.error(
                f"Invalid log out request for user {username}")
        jwk_id: Optional[str] = self.users[username].jwk_id
        self.users[username].jwt_secret = None
        self.users[username].jwk_id = None
        if jwk_id is not None:
            self.public_jwks.pop(jwk_id, None)
        await self._sync_user(username)
        eventloop = self.server.get_event_loop()
        eventloop.delay_callback(
            .005, self.server.send_event, "authorization:user_logged_out",
            {'username': username}
        )
        return {
            "username": username,
            "action": "user_logged_out"
        }

    async def _handle_info_request(self, web_request: WebRequest) -> Dict[str, Any]:
        sources = ["moonraker"]
        if self.ldap is not None:
            sources.append("ldap")
        login_req = self.force_logins and len(self.users) > 1
        request_trusted: Optional[bool] = None
        user = web_request.get_current_user()
        req_ip = web_request.ip_addr
        if user is not None and user.username == TRUSTED_USER:
            request_trusted = True
        elif req_ip is not None:
            request_trusted = await self._check_authorized_ip(req_ip)
        return {
            "default_source": self.default_source,
            "available_sources": sources,
            "login_required": login_req,
            "trusted": request_trusted
        }

    async def _handle_refresh_jwt(self,
                                  web_request: WebRequest
                                  ) -> Dict[str, str]:
        refresh_token: str = web_request.get_str('refresh_token')
        try:
            user_info = self.decode_jwt(refresh_token, token_type="refresh")
        except Exception:
            raise self.server.error("Invalid Refresh Token", 401)
        username: str = user_info.username
        if user_info.jwt_secret is None or user_info.jwk_id is None:
            raise self.server.error("User not logged in", 401)
        private_key = self._load_private_key(user_info.jwt_secret)
        jwk_id: str = user_info.jwk_id
        token = self._generate_jwt(username, jwk_id, private_key)
        return {
            'username': username,
            'token': token,
            'source': user_info.source,
            'action': 'user_jwt_refresh'
        }

    async def _handle_user_request(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        req_type = web_request.get_request_type()
        if req_type == RequestType.GET:
            user = web_request.get_current_user()
            if user is None:
                return {
                    "username": None,
                    "source": None,
                    "created_on": None,
                }
            else:
                return {
                    "username": user.username,
                    "source": user.source,
                    "created_on": user.created_on
                }
        elif req_type == RequestType.POST:
            # Create User
            return await self._login_jwt_user(web_request, create=True)
        elif req_type == RequestType.DELETE:
            # Delete User
            return await self._delete_jwt_user(web_request)
        raise self.server.error("Invalid Request Method")

    async def _handle_list_request(self,
                                   web_request: WebRequest
                                   ) -> Dict[str, List[Dict[str, Any]]]:
        user_list = []
        for user in self.users.values():
            if user.username == API_USER:
                continue
            user_list.append({
                'username': user.username,
                'source': user.source,
                'created_on': user.created_on
            })
        return {
            'users': user_list
        }

    async def _handle_password_reset(self,
                                     web_request: WebRequest
                                     ) -> Dict[str, str]:
        password: str = web_request.get_str('password')
        new_pass: str = web_request.get_str('new_password')
        user_info = web_request.get_current_user()
        if user_info is None:
            raise self.server.error("No Current User")
        username = user_info.username
        if user_info.source == "ldap":
            raise self.server.error(
                f"Can´t Reset password for ldap user {username}")
        if username in RESERVED_USERS:
            raise self.server.error(
                f"Invalid Reset Request for user {username}")
        salt = bytes.fromhex(user_info.salt)
        hashed_pass = hashlib.pbkdf2_hmac(
            'sha256', password.encode(), salt, HASH_ITER).hex()
        if hashed_pass != user_info.password:
            raise self.server.error("Invalid Password")
        new_hashed_pass = hashlib.pbkdf2_hmac(
            'sha256', new_pass.encode(), salt, HASH_ITER).hex()
        self.users[username].password = new_hashed_pass
        await self._sync_user(username)
        return {
            'username': username,
            'action': "user_password_reset"
        }

    async def _login_jwt_user(
        self, web_request: WebRequest, create: bool = False
    ) -> Dict[str, Any]:
        username: str = web_request.get_str('username')
        password: str = web_request.get_str('password')
        source: str = web_request.get_str(
            'source', self.default_source
        ).lower()
        if source not in AUTH_SOURCES:
            raise self.server.error(f"Invalid 'source': {source}")
        user_info: UserInfo
        if username in RESERVED_USERS:
            raise self.server.error(
                f"Invalid Request for user {username}")
        if source == "ldap":
            if create:
                raise self.server.error("Cannot Create LDAP User")
            if self.ldap is None:
                raise self.server.error(
                    "LDAP authentication not available", 401
                )
            await self.ldap.authenticate_ldap_user(username, password)
            if username not in self.users:
                create = True
        if create:
            if username in self.users:
                raise self.server.error(f"User {username} already exists")
            salt = secrets.token_bytes(32)
            hashed_pass = hashlib.pbkdf2_hmac(
                'sha256', password.encode(), salt, HASH_ITER).hex()
            user_info = UserInfo(
                username=username,
                password=hashed_pass,
                salt=salt.hex(),
                source=source,
            )
            self.users[username] = user_info
            await self._sync_user(username)
            action = "user_created"
            if source == "ldap":
                # Dont notify user created
                action = "user_logged_in"
                create = False
        else:
            if username not in self.users:
                raise self.server.error(f"Unregistered User: {username}")
            user_info = self.users[username]
            auth_src = user_info.source
            if auth_src != source:
                raise self.server.error(
                    f"Moonraker cannot authenticate user '{username}', must "
                    f"specify source '{auth_src}'", 401
                )
            salt = bytes.fromhex(user_info.salt)
            hashed_pass = hashlib.pbkdf2_hmac(
                'sha256', password.encode(), salt, HASH_ITER).hex()
            action = "user_logged_in"
            if hashed_pass != user_info.password:
                raise self.server.error("Invalid Password")
        jwt_secret_hex: Optional[str] = user_info.jwt_secret
        if jwt_secret_hex is None:
            private_key = Signer()
            jwk_id = base64url_encode(secrets.token_bytes()).decode()
            user_info.jwt_secret = private_key.hex_seed().decode()  # type: ignore
            user_info.jwk_id = jwk_id
            self.users[username] = user_info
            await self._sync_user(username)
            self.public_jwks[jwk_id] = self._generate_public_jwk(private_key)
        else:
            private_key = self._load_private_key(jwt_secret_hex)
            if user_info.jwk_id is None:
                user_info.jwk_id = base64url_encode(secrets.token_bytes()).decode()
            jwk_id = user_info.jwk_id
        token = self._generate_jwt(username, jwk_id, private_key)
        refresh_token = self._generate_jwt(
            username, jwk_id, private_key, token_type="refresh",
            exp_time=datetime.timedelta(days=self.login_timeout))
        conn = web_request.get_client_connection()
        if create:
            event_loop = self.server.get_event_loop()
            event_loop.delay_callback(
                .005, self.server.send_event,
                "authorization:user_created",
                {'username': username})
        elif conn is not None:
            conn.user_info = user_info
        return {
            'username': username,
            'token': token,
            'source': user_info.source,
            'refresh_token': refresh_token,
            'action': action
        }

    async def _delete_jwt_user(self, web_request: WebRequest) -> Dict[str, str]:
        username: str = web_request.get_str('username')
        current_user = web_request.get_current_user()
        if current_user is not None:
            curname = current_user.username
            if curname == username:
                raise self.server.error(f"Cannot delete logged in user {curname}")
        if username in RESERVED_USERS:
            raise self.server.error(
                f"Invalid Request for reserved user {username}")
        user_info: Optional[UserInfo] = self.users.get(username)
        if user_info is None:
            raise self.server.error(f"No registered user: {username}")
        if user_info.jwk_id is not None:
            self.public_jwks.pop(user_info.jwk_id, None)
        del self.users[username]
        async with self.user_table as tx:
            await tx.execute(
                f"DELETE FROM {USER_TABLE} WHERE username = ?", (username,)
            )
        event_loop = self.server.get_event_loop()
        event_loop.delay_callback(
            .005, self.server.send_event,
            "authorization:user_deleted",
            {'username': username})
        return {
            "username": username,
            "action": "user_deleted"
        }

    def _generate_jwt(self,
                      username: str,
                      jwk_id: str,
                      private_key: Signer,
                      token_type: str = "access",
                      exp_time: datetime.timedelta = JWT_EXP_TIME
                      ) -> str:
        curtime = int(time.time())
        payload = {
            'iss': self.issuer,
            'aud': "Moonraker",
            'iat': curtime,
            'exp': curtime + int(exp_time.total_seconds()),
            'username': username,
            'token_type': token_type
        }
        header = {'kid': jwk_id}
        header.update(JWT_HEADER)
        jwt_header = base64url_encode(jsonw.dumps(header))
        jwt_payload = base64url_encode(jsonw.dumps(payload))
        jwt_msg = b".".join([jwt_header, jwt_payload])
        sig = private_key.signature(jwt_msg)
        jwt_sig = base64url_encode(sig)
        return b".".join([jwt_msg, jwt_sig]).decode()

    def decode_jwt(
        self, token: str, token_type: str = "access", check_exp: bool = True
    ) -> UserInfo:
        message, sig = token.rsplit('.', maxsplit=1)
        enc_header, enc_payload = message.split('.')
        header: Dict[str, Any] = jsonw.loads(base64url_decode(enc_header))
        sig_bytes = base64url_decode(sig)

        # verify header
        if header.get('typ') != "JWT" or header.get('alg') != "EdDSA":
            raise self.server.error("Invalid JWT header")
        jwk_id: Optional[str] = header.get('kid')
        if jwk_id not in self.public_jwks:
            raise self.server.error("Invalid key ID")

        # validate signature
        public_key = self._public_key_from_jwk(self.public_jwks[jwk_id])
        public_key.verify(sig_bytes + message.encode())

        # validate claims
        payload: Dict[str, Any] = jsonw.loads(base64url_decode(enc_payload))
        if payload['token_type'] != token_type:
            raise self.server.error(
                f"JWT Token type mismatch: Expected {token_type}, "
                f"Recd: {payload['token_type']}", 401)
        if payload['iss'] != self.issuer:
            raise self.server.error("Invalid JWT Issuer", 401)
        if payload['aud'] != "Moonraker":
            raise self.server.error("Invalid JWT Audience", 401)
        if check_exp and payload['exp'] < int(time.time()):
            raise self.server.error("JWT Expired", 401)

        # get user
        user_info: Optional[UserInfo] = self.users.get(
            payload.get('username', ""), None)
        if user_info is None:
            raise self.server.error("Unknown user", 401)
        return user_info

    def validate_jwt(self, token: str) -> UserInfo:
        try:
            user_info = self.decode_jwt(token)
        except Exception as e:
            if isinstance(e, self.server.error):
                raise
            raise self.server.error(
                f"Failed to decode JWT: {e}", 401
            ) from e
        return user_info

    def validate_api_key(self, api_key: str) -> UserInfo:
        if not self.enable_api_key:
            raise self.server.error("API Key authentication is disabled", 401)
        if api_key and api_key == self.api_key:
            return self.users[API_USER]
        raise self.server.error("Invalid API Key", 401)

    def _load_private_key(self, secret: str) -> Signer:
        try:
            key = Signer(bytes.fromhex(secret))
        except Exception:
            raise self.server.error(
                "Error decoding private key, user data may"
                " be corrupt", 500) from None
        return key

    def _generate_public_jwk(self, private_key: Signer) -> Dict[str, Any]:
        public_key = private_key.vk
        return {
            'x': base64url_encode(public_key).decode(),
            'kty': "OKP",
            'crv': "Ed25519",
            'use': "sig"
        }

    def _public_key_from_jwk(self, jwk: Dict[str, Any]) -> Verifier:
        if jwk.get('kty') != "OKP":
            raise self.server.error("Not an Octet Key Pair")
        if jwk.get('crv') != "Ed25519":
            raise self.server.error("Invalid Curve")
        if 'x' not in jwk:
            raise self.server.error("No 'x' argument in jwk")
        key = base64url_decode(jwk['x'])
        return Verifier(key.hex().encode())

    def _prune_conn_handler(self, eventtime: float) -> float:
        cur_time = time.time()
        for ip, user_info in list(self.trusted_users.items()):
            exp_time: float = user_info['expires_at']
            if cur_time >= exp_time:
                self.trusted_users.pop(ip, None)
                logging.info(f"Trusted Connection Expired, IP: {ip}")
        for ip, fqdn_info in list(self.fqdn_cache.items()):
            exp_time = fqdn_info["expires_at"]
            if cur_time >= exp_time:
                domain: str = fqdn_info["domain"]
                self.fqdn_cache.pop(ip, None)
                logging.info(f"Cached FQDN Expired, IP: {ip}, domain: {domain}")
        return eventtime + PRUNE_CHECK_TIME

    def _oneshot_token_expire_handler(self, token):
        self.oneshot_tokens.pop(token, None)

    def get_oneshot_token(self, ip_addr: IPAddr, user: Optional[UserInfo]) -> str:
        token = base64.b32encode(os.urandom(20)).decode()
        event_loop = self.server.get_event_loop()
        hdl = event_loop.delay_callback(
            ONESHOT_TIMEOUT, self._oneshot_token_expire_handler, token)
        self.oneshot_tokens[token] = (ip_addr, user, hdl)
        return token

    def _check_json_web_token(
        self, request: HTTPServerRequest, required: bool = True
    ) -> Optional[UserInfo]:
        auth_token: Optional[str] = request.headers.get("Authorization")
        if auth_token is None:
            auth_token = request.headers.get("X-Access-Token")
            if auth_token is None:
                qtoken = request.query_arguments.get('access_token', None)
                if qtoken is not None:
                    auth_token = qtoken[-1].decode(errors="ignore")
        elif auth_token.startswith("Bearer "):
            auth_token = auth_token[7:]
        else:
            return None
        if auth_token:
            try:
                return self.decode_jwt(auth_token, check_exp=required)
            except Exception:
                logging.exception(f"JWT Decode Error {auth_token}")
                raise HTTPError(401, "JWT Decode Error")
        return None

    async def _check_authorized_ip(self, ip: IPAddr) -> bool:
        if ip in self.trusted_ips:
            return True
        for rng in self.trusted_ranges:
            if ip in rng:
                return True
        if self.trusted_domains:
            if ip in self.fqdn_cache:
                fqdn: str = self.fqdn_cache[ip]["domain"]
            else:
                eventloop = self.server.get_event_loop()
                try:
                    fut = eventloop.run_in_thread(socket.getfqdn, str(ip))
                    fqdn = await asyncio.wait_for(fut, 5.0)
                except asyncio.TimeoutError:
                    logging.info("Call to socket.getfqdn() timed out")
                    return False
                else:
                    fqdn = fqdn.lower()
                    self.fqdn_cache[ip] = {
                        "expires_at": time.time() + FQDN_CACHE_TIMEOUT,
                        "domain": fqdn
                    }
            return fqdn in self.trusted_domains
        return False

    async def _check_trusted_connection(
        self, ip: Optional[IPAddr]
    ) -> Optional[UserInfo]:
        if ip is not None:
            curtime = time.time()
            exp_time = curtime + TRUSTED_CONNECTION_TIMEOUT
            if ip in self.trusted_users:
                self.trusted_users[ip]["expires_at"] = exp_time
                return self.trusted_users[ip]["user"]
            elif await self._check_authorized_ip(ip):
                logging.info(
                    f"Trusted Connection Detected, IP: {ip}")
                self.trusted_users[ip] = {
                    "user": UserInfo(TRUSTED_USER, "", curtime),
                    "expires_at": exp_time
                }
                return self.trusted_users[ip]["user"]
        return None

    def _check_oneshot_token(
        self, token: str, cur_ip: Optional[IPAddr]
    ) -> Optional[UserInfo]:
        if token in self.oneshot_tokens:
            ip_addr, user, hdl = self.oneshot_tokens.pop(token)
            hdl.cancel()
            if cur_ip != ip_addr:
                logging.info(f"Oneshot Token IP Mismatch: expected{ip_addr}"
                             f", Recd: {cur_ip}")
                return None
            return user
        else:
            return None

    def check_logins_maxed(self, ip_addr: IPAddr) -> bool:
        if self.max_logins is None:
            return False
        return self.failed_logins.get(ip_addr, 0) >= self.max_logins

    async def authenticate_request(
        self, request: HTTPServerRequest, auth_required: bool = True
    ) -> Optional[UserInfo]:
        if request.method == "OPTIONS":
            return None

        # Check JSON Web Token
        jwt_user = self._check_json_web_token(request, auth_required)
        if jwt_user is not None:
            return jwt_user

        try:
            ip = ipaddress.ip_address(request.remote_ip)  # type: ignore
        except ValueError:
            logging.exception(
                f"Unable to Create IP Address {request.remote_ip}")
            ip = None

        # Check oneshot access token
        ost: Optional[List[bytes]] = request.arguments.get('token', None)
        if ost is not None:
            ost_user = self._check_oneshot_token(ost[-1].decode(), ip)
            if ost_user is not None:
                return ost_user

        # Check API Key Header
        if self.enable_api_key:
            key: Optional[str] = request.headers.get("X-Api-Key")
            if key and key == self.api_key:
                return self.users[API_USER]

        # If the force_logins option is enabled and at least one user is created
        # then trusted user authentication is disabled
        if self.force_logins and len(self.users) > 1:
            if not auth_required:
                return None
            raise HTTPError(401, "Unauthorized, Force Logins Enabled")

        # Check if IP is trusted.  If this endpoint doesn't require authentication
        # then it is acceptable to return None
        trusted_user = await self._check_trusted_connection(ip)
        if trusted_user is not None:
            return trusted_user
        if not auth_required:
            return None

        raise HTTPError(401, "Unauthorized")

    async def check_cors(self, origin: Optional[str]) -> bool:
        if origin is None or not self.cors_domains:
            return False
        for regex in self.cors_domains:
            match = re.match(regex, origin)
            if match is not None:
                if match.group() == origin:
                    logging.debug(f"CORS Pattern Matched, origin: {origin} "
                                  f" | pattern: {regex}")
                    return True
                else:
                    logging.debug(f"Partial Cors Match: {match.group()}")
        else:
            # Check to see if the origin contains an IP that matches a
            # current trusted connection
            match = re.search(r"^https?://([^/:]+)", origin)
            if match is not None:
                ip = match.group(1)
                try:
                    ipaddr = ipaddress.ip_address(ip)
                except ValueError:
                    pass
                else:
                    if await self._check_authorized_ip(ipaddr):
                        logging.debug(f"Cors request matched trusted IP: {ip}")
                        return True
            logging.debug(f"No CORS match for origin: {origin}\n"
                          f"Patterns: {self.cors_domains}")
        return False

    def cors_enabled(self) -> bool:
        return self.cors_domains is not None

    def close(self) -> None:
        self.prune_timer.stop()


def load_component(config: ConfigHelper) -> Authorization:
    return Authorization(config)
